# Copyright (c) lobsterpy development team
# Distributed under the terms of a BSD 3-Clause "New" or "Revised" License

"""This module defines classes to analyze the COHPs automatically."""

from __future__ import annotations

import warnings
from collections import Counter
from pathlib import Path

import numpy as np
from pymatgen.analysis.bond_valence import BVAnalyzer
from pymatgen.core.structure import Structure
from pymatgen.electronic_structure.cohp import CompleteCohp
from pymatgen.electronic_structure.core import Spin
from pymatgen.electronic_structure.dos import LobsterCompleteDos
from pymatgen.io.lobster import (
    Bandoverlaps,
    Charge,
    Doscar,
    Icohplist,
    Lobsterin,
    Lobsterout,
    MadelungEnergies,
)
from pymatgen.io.lobster.lobsterenv import LobsterNeighbors
from pymatgen.io.vasp.outputs import Vasprun
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer


class Analysis:
    """
    Class to automatically analyze COHP/COOP/COBI  populations from Lobster.

    :param are_cobis: bool indicating if file contains COBI/ICOBI data
    :param are_coops: bool indicating if file contains COOP/ICOOP data
    :param cutoff_icohp: Cutoff in percentage for evaluating neighbors based on ICOHP values.
        cutoff_icohp*max_icohp limits the number of considered neighbours for evaluating environments.
    :param path_to_cohpcar: path to `COHPCAR.lobster` or `COBICAR.lobster` or `COOPCAR.lobster` .
    :param path_to_charge: path to `CHARGE.lobster`.
    :param path_to_icohplist: path to `ICOHPLIST.lobster` or `ICOBILIST.lobster` or `ICOOPLIST.lobster`.
    :param path_to_poscar: path to structure (e.g., `POSCAR` or `POSCAR.lobster`)
    :param path_to_madelung: path to `MadelungEnergies.lobster`.
    :param charge_obj: pymatgen lobster.io.charge object
    :param completecohp_obj: pymatgen.electronic_structure.cohp.CompleteCohp object
    :param icohplist_obj: pymatgen lobster.io.Icohplist object
    :param madelung_obj: pymatgen lobster.io.MadelungEnergies object
    :param noise_cutoff: Sets the lower limit tolerance for ICOHPs or ICOOPs or ICOBIs considered
        in analysis.
    :param orbital_cutoff: Sets the minimum percentage for the orbital contribution considered to be
        relevant in orbital resolved analysis. (Affects only when orbital_resolved argument is set to True)
        Set it to 0 to get results of all orbitals in the detected relevant bonds. Default is to 0.05 i.e.
        only analyzes if orbital contribution is 5 % or more.
    :param orbital_resolved: bool indicating whether orbital wise analysis is performed
    :param type_charge: If no path_to_charge is given, Valences will be used. Otherwise, Mulliken charges.
            Löwdin charges cannot be selected at the moment.
    :param which_bonds: Selects kinds of bonds that are analyzed. `cation-anion` is the default.
        Alternatively, `all` bonds can also be selected. Support to other kinds of bonds will be
        added soon.
    :param summed_spins: if True, COHP `Spin.up` and `Spin.down` populations will be summed
    :param start: sets the lower limit of energy for evaluation of bonding and antibonding
        percentages below efermi. Defaults to None (i.e., all populations below efermi are included)

    Attributes:
        - condensed_bonding_analysis: dict including a summary of the most important bonding properties
        - final_dict_bonds: dict including information on ICOHPs per bond type
        - final_dict_ions: dict including information on environments of cations
        - chemenv: pymatgen.io.lobster.lobsterenv.LobsterNeighbors object
        - lse: LightStructureEnvironment from pymatgen
        - anion_types: Set of Element objects from pymatgen
        - list_equivalent_sites: list of site indices of sites that indicate which sites are equivalent
            e.g., [0 1 2 2 2] where site 0, 1, 2 indicate sites that are independent from each other
        - seq_cohps: list of cohps
        - seq_coord_ions: list of co-ordination environment strings for each cation
        - seq_equivalent_sites: seq of inequivalent sites
        - seq_ineq_ions: seq of inequivalent cations/sites in the structure
        - seq_infos_bonds (list): information on cation anion bonds (lists
            of pymatgen.io.lobster.lobsterenv.ICOHPNeighborsInfo)
        - spg: space group information
        - structure: Structure object

    """

    def __init__(
        self,
        path_to_poscar: str | Path | None,
        path_to_icohplist: str | Path | None,
        path_to_cohpcar: str | Path | None,
        path_to_charge: str | Path | None = None,
        path_to_madelung: str | Path | None = None,
        icohplist_obj: Icohplist | None = None,
        completecohp_obj: CompleteCohp | None = None,
        charge_obj: Charge | None = None,
        madelung_obj: MadelungEnergies | None = None,
        are_cobis: bool = False,
        are_coops: bool = False,
        cutoff_icohp: float = 0.1,
        noise_cutoff: float = 0.1,
        orbital_cutoff: float = 0.05,
        orbital_resolved: bool = False,
        start: float | None = None,
        summed_spins: bool = True,
        type_charge: str | None = None,
        which_bonds: str = "cation-anion",
    ):
        """
        Initialize automatic bonding analysis.

        :param are_cobis: bool indicating if file contains COBI/ICOBI data
        :param are_coops: bool indicating if file contains COOP/ICOOP data
        :param cutoff_icohp: Cutoff in percentage for evaluating neighbors based on ICOHP values.
            cutoff_icohp*max_icohp limits the number of considered neighbours for evaluating environments.
        :param path_to_cohpcar: path to `COHPCAR.lobster` or `COBICAR.lobster` or `COOPCAR.lobster` .
        :param path_to_charge: path to `CHARGE.lobster`.
        :param path_to_icohplist: path to `ICOHPLIST.lobster` or `ICOBILIST.lobster` or `ICOOPLIST.lobster`.
        :param path_to_poscar: path to structure (e.g., `POSCAR` or `POSCAR.lobster`)
        :param path_to_madelung: path to `MadelungEnergies.lobster`.
        :param charge_obj: pymatgen lobster.io.charge object (Optional)
        :param completecohp_obj: pymatgen.electronic_structure.cohp.CompleteCohp object
        :param icohplist_obj: pymatgen lobster.io.Icohplist object
        :param madelung_obj: pymatgen lobster.io.MadelungEnergies object
        :param noise_cutoff: Sets the lower limit tolerance for ICOHPs or ICOOPs or ICOBIs considered
            in analysis.
        :param orbital_cutoff: Sets the minimum percentage for the orbital contribution considered to be
            relevant in orbital resolved analysis. (Affects only when orbital_resolved argument is set to True)
            Set it to 0 to get results of all orbitals in the detected relevant bonds. Default is to 0.05 i.e.
            only analyzes if orbital contribution is 5 % or more.
        :param orbital_resolved: bool indicating whether orbital wise analysis is performed
        :param type_charge: If no path_to_charge is given, Valences will be used. Otherwise, Mulliken charges.
                Löwdin charges cannot be selected at the moment.
        :param which_bonds: Selects kinds of bonds that are analyzed. `cation-anion` is the default.
            Alternatively, `all` bonds can also be selected. Support to other kinds of bonds will be
            added soon.
        :param summed_spins: if True, COHP `Spin.up` and `Spin.down` populations will be summed
        :param start: sets the lower limit of energy for evaluation of bonding and antibonding
            percentages below efermi. Defaults to None (i.e., all populations below efermi are included)

        """
        self.start = start
        self.completecohp_obj = completecohp_obj
        self.icohplist_obj = icohplist_obj
        # checks to ensure LobsterEnv inputs are not duplicated in case users provide both path and obj
        if self.completecohp_obj is not None and self.icohplist_obj is not None:
            self.path_to_poscar = None
            self.path_to_cohpcar = None
            self.path_to_icohplist = None
        else:
            self.path_to_poscar = path_to_poscar
            self.path_to_icohplist = path_to_icohplist
            self.path_to_cohpcar = path_to_cohpcar
        self.which_bonds = which_bonds
        self.cutoff_icohp = cutoff_icohp
        self.orbital_cutoff = orbital_cutoff
        self.path_to_charge = path_to_charge
        self.charge_obj = charge_obj
        self.path_to_madelung = path_to_madelung
        self.madelung_obj = madelung_obj
        self.are_cobis = are_cobis
        self.are_coops = are_coops
        self.noise_cutoff = noise_cutoff
        self.setup_env()
        self.get_information_all_bonds(summed_spins=summed_spins)
        self.orbital_resolved = orbital_resolved

        # This determines how cations and anions
        if path_to_charge is None and charge_obj is None:
            self.type_charge = "Valences"
        else:
            if type_charge is None:
                self.type_charge = "Mulliken"
            elif type_charge == "Mulliken":
                self.type_charge = "Mulliken"
            elif type_charge == "Löwdin":
                raise ValueError("Only Mulliken charges can be used here at the moment. Implementation will follow.")
            else:
                self.type_charge = "Valences"
                print("type_charge cannot be read! Please use Mulliken/Löwdin. Now, we will use valences")

        self.set_condensed_bonding_analysis()
        self.set_summary_dicts()
        self.path_to_madelung = path_to_madelung

    def setup_env(self):
        """
        Set up the light structure environments based on COHPs using this method.

        Returns:
            None

        """
        self.structure = (
            Structure.from_file(self.path_to_poscar) if self.path_to_poscar else self.completecohp_obj.structure
        )
        sga = SpacegroupAnalyzer(structure=self.structure)
        symmetry_dataset = sga.get_symmetry_dataset()
        equivalent_sites = symmetry_dataset["equivalent_atoms"]
        self.list_equivalent_sites = equivalent_sites
        self.seq_equivalent_sites = list(set(equivalent_sites))
        self.spg = symmetry_dataset["international"]

        if self.which_bonds == "cation-anion":
            try:
                self.chemenv = LobsterNeighbors(
                    filename_icohp=self.path_to_icohplist,
                    obj_icohp=self.icohplist_obj,
                    structure=self.structure,
                    additional_condition=1,
                    perc_strength_icohp=self.cutoff_icohp,
                    filename_charge=self.path_to_charge,
                    obj_charge=self.charge_obj,
                    valences=None,
                    valences_from_charges=True,
                    adapt_extremum_to_add_cond=True,
                    are_cobis=self.are_cobis,
                    are_coops=self.are_coops,
                    noise_cutoff=self.noise_cutoff,
                )
            except ValueError as err:
                if (
                    str(err) == "min() arg is an empty sequence"
                    or str(err) == "All valences are equal to 0, additional_conditions 1, 3, 5 and 6 will not work"
                ):
                    raise ValueError(
                        "Consider switching to an analysis of all bonds and not only cation-anion bonds."
                        " It looks like no cations are detected."
                    )
                raise err
        elif self.which_bonds == "all":
            # raise ValueError("only cation anion bonds implemented so far")
            self.chemenv = LobsterNeighbors(
                filename_icohp=self.path_to_icohplist,
                obj_icohp=self.icohplist_obj,
                structure=self.structure,
                additional_condition=0,
                perc_strength_icohp=self.cutoff_icohp,
                filename_charge=self.path_to_charge,
                obj_charge=self.charge_obj,
                valences_from_charges=True,
                adapt_extremum_to_add_cond=True,
                are_cobis=self.are_cobis,
                are_coops=self.are_coops,
                noise_cutoff=self.noise_cutoff,
            )

        else:
            raise ValueError("only cation anion and all bonds implemented so far")

        # determine cations and anions
        try:
            if self.which_bonds == "cation-anion":
                self.lse = self.chemenv.get_light_structure_environment(only_cation_environments=True)
            elif self.which_bonds == "all":
                self.lse = self.chemenv.get_light_structure_environment(only_cation_environments=False)
        except ValueError:

            class Lse:
                """Test class when error was raised."""

                def __init__(self, chemenv, valences=None):
                    """
                    Test class when error was raised.

                    :param chemenv: LobsterNeighbors object
                    :param valences: list of valences

                    """
                    if valences is None:
                        self.coordination_environments = []
                        for coord in chemenv:
                            if len(coord) > 0:
                                self.coordination_environments.append([{"ce_symbol": str(len(coord))}])
                            else:
                                self.coordination_environments.append([{"ce_symbol": None}])
                    else:
                        self.coordination_environments = []

                        for val, coord in zip(valences, chemenv):
                            if val >= 0.0 and len(coord) > 0:
                                self.coordination_environments.append([{"ce_symbol": str(len(coord))}])
                            else:
                                self.coordination_environments.append([{"ce_symbol": None}])

            if self.which_bonds == "all":
                self.lse = Lse(self.chemenv.list_coords)
            elif self.which_bonds == "cation-anion":
                # make a new list
                self.lse = Lse(self.chemenv.list_coords, self.chemenv.valences)

    def get_information_all_bonds(self, summed_spins: bool = True):
        """
        Gather all information on the bonds within the compound with this method.

        Returns:
            None

        """
        if self.which_bonds == "cation-anion":
            # this will only analyze cation anion bonds which simplifies the analysis
            self.seq_ineq_ions = []
            self.seq_coord_ions = []
            self.seq_infos_bonds = []
            self.seq_labels_cohps = []
            self.seq_cohps = []
            # only_bonds_to

            self.anion_types = self.chemenv.anion_types
            for ice, ce in enumerate(self.lse.coordination_environments):
                # only look at inequivalent sites (use of symmetry to speed everything up!)!
                # only look at those cations that have cation-anion bonds
                if ice in self.seq_equivalent_sites and ce[0]["ce_symbol"] is not None:
                    self.seq_ineq_ions.append(ice)

                    self.seq_coord_ions.append(ce[0]["ce_symbol"])
                    self.seq_infos_bonds.append(self.chemenv.get_info_icohps_to_neighbors([ice]))

                    aniontype_labels = []
                    aniontype_cohps = []

                    # go through all anions in the structure!
                    for anion in self.anion_types:
                        # get labels and summed cohp objects
                        labels, summedcohps = self.chemenv.get_info_cohps_to_neighbors(
                            path_to_cohpcar=self.path_to_cohpcar,
                            obj_cohpcar=self.completecohp_obj,
                            isites=[ice],
                            summed_spin_channels=summed_spins,
                            per_bond=False,
                            only_bonds_to=[str(anion)],
                        )

                        aniontype_labels.append(labels)
                        aniontype_cohps.append(summedcohps)

                    self.seq_labels_cohps.append(aniontype_labels)
                    self.seq_cohps.append(aniontype_cohps)

        elif self.which_bonds == "all":
            # this will only analyze all bonds

            self.seq_ineq_ions = []
            self.seq_coord_ions = []
            self.seq_infos_bonds = []
            self.seq_labels_cohps = []
            self.seq_cohps = []
            # only_bonds_to
            self.elements = self.structure.composition.elements
            # self.anion_types = self.chemenv.get_anion_types()
            for ice, ce in enumerate(self.lse.coordination_environments):
                # only look at inequivalent sites (use of symmetry to speed everything up!)!
                # only look at those cations that have cation-anion bonds
                if ice in self.seq_equivalent_sites and ce[0]["ce_symbol"] is not None:
                    self.seq_ineq_ions.append(ice)
                    self.seq_coord_ions.append(ce[0]["ce_symbol"])
                    self.seq_infos_bonds.append(self.chemenv.get_info_icohps_to_neighbors([ice]))

                    type_labels = []
                    type_cohps = []

                    for element in self.elements:
                        # get labels and summed cohp objects
                        labels, summedcohps = self.chemenv.get_info_cohps_to_neighbors(
                            path_to_cohpcar=self.path_to_cohpcar,
                            obj_cohpcar=self.completecohp_obj,
                            isites=[ice],
                            onlycation_isites=False,
                            summed_spin_channels=summed_spins,
                            per_bond=False,
                            only_bonds_to=[str(element)],
                        )

                        type_labels.append(labels)
                        type_cohps.append(summedcohps)

                    self.seq_labels_cohps.append(type_labels)
                    self.seq_cohps.append(type_cohps)

    def get_site_bond_resolved_labels(self):
        """
        Return relevant bond labels for each symmetrically independent site.

        Returns:
            dict with bond labels for each site, e.g.
            {'Na1: Na-Cl': ['21', '23', '24', '27', '28', '30']}

        """
        bonds = [[] for _ in range(len(self.seq_infos_bonds))]  # type: ignore
        labels = [[] for _ in range(len(self.seq_infos_bonds))]  # type: ignore
        for inx, bond_info in enumerate(self.seq_infos_bonds):
            for ixx, val in enumerate(bond_info.atoms):
                label_srt = sorted(val.copy())
                bonds[inx].append(
                    self.structure.sites[bond_info.central_isites[0]].species_string
                    + str(bond_info.central_isites[0] + 1)
                    + ": "
                    + label_srt[0].strip("0123456789")
                    + "-"
                    + label_srt[1].strip("0123456789")
                )
                labels[inx].append(bond_info.labels[ixx])

        label_data = {}
        for indx, atom_pairs in enumerate(bonds):
            searched_atom_pairs = set(atom_pairs)
            for search_item in searched_atom_pairs:
                indices = [i for i, pair in enumerate(atom_pairs) if pair == search_item]
                filtered_bond_label_list = [labels[indx][i] for i in indices]
                label_data.update({search_item: filtered_bond_label_list})

        return label_data

    def _get_orbital_resolved_data(
        self,
        nameion: str,
        iion: int,
        labels: list[str],
        bond_resolved_labels: dict[str, list[str]],
        type_pop: str,
    ):
        """
        Retrieve orbital-wise analysis data.

        :param nameion: name of symmetrically relevant cation or anion
        :param iion: index of symmetrically relevant cation or anion
        :param labels: list of bond label names
        :param bond_resolved_labels: dict of bond labels from ICOHPLIST resolved for each bond
        :param type_pop: population type analyzed. e.g. COHP or COOP or COBI

        Returns:
            dict consisting of relevant orbitals (contribution > 5 % to overall ICOHP or ICOBI or ICOOP),
            bonding and antibonding percentages with bond label names as keys.
        """
        orb_resolved_bond_info = {}
        for label in labels:
            if label is not None:
                bond_resolved_label_key = nameion + str(iion + 1) + ":" + label.split("x")[-1]
                bond_labels = bond_resolved_labels[bond_resolved_label_key]
                available_orbitals = list(self.chemenv.completecohp.orb_res_cohp[bond_labels[0]].keys())
                # initialize empty list to store orb paris for bonding,
                # antibonding integrals and  percentages
                bndg_orb_pair_list = []
                bndg_orb_integral_list = []
                bndg_orb_perc_list = []
                bndg_orb_icohp_list = []
                antibndg_orb_integral_list = []
                antibndg_orb_perc_list = []
                antibndg_orb_pair_list = []
                antibndg_orb_icohp_list = []

                # get total summed cohps using label list
                cohp_summed = self.chemenv.completecohp.get_summed_cohp_by_label_list(label_list=bond_labels)
                if type_pop.lower() == "cohp":
                    (
                        antibndg_tot,
                        per_anti_tot,
                        bndg_tot,
                        per_bndg_tot,
                    ) = self._integrate_antbdstates_below_efermi(cohp=cohp_summed, start=self.start)
                else:
                    (
                        bndg_tot,
                        per_bndg_tot,
                        antibndg_tot,
                        per_anti_tot,
                    ) = self._integrate_antbdstates_below_efermi(cohp=cohp_summed, start=self.start)

                orb_bonding_dict_data = {}  # type: ignore
                # For each orbital collect the contributions of summed bonding
                # and antibonding interactions separately
                for orb in available_orbitals:
                    cohp_summed_orb = self.chemenv.completecohp.get_summed_cohp_by_label_and_orbital_list(
                        label_list=bond_labels, orbital_list=[orb] * len(bond_labels)
                    )

                    if type_pop.lower() == "cohp":
                        (
                            antibndg_orb,
                            per_anti_orb,
                            bndg_orb,
                            per_bndg_orb,
                        ) = self._integrate_antbdstates_below_efermi(cohp=cohp_summed_orb, start=self.start)
                    else:
                        (
                            bndg_orb,
                            per_bndg_orb,
                            antibndg_orb,
                            per_anti_orb,
                        ) = self._integrate_antbdstates_below_efermi(cohp=cohp_summed_orb, start=self.start)

                    # replace nan values with zero (tackle numerical integration issues)
                    bndg_orb = bndg_orb if not np.isnan(bndg_orb) else 0
                    per_bndg_orb = per_bndg_orb if not np.isnan(per_bndg_orb) else 0
                    bndg_tot = bndg_tot if not np.isnan(bndg_tot) else 0
                    per_bndg_tot = per_bndg_tot if not np.isnan(per_bndg_tot) else 0
                    # skip collecting orb contributions if no summed bonding contribution exists
                    if bndg_tot > 0:
                        orb_icohps_bndg = []
                        for bond_label in bond_labels:
                            orb_icohp_bn = self.chemenv.Icohpcollection.get_icohp_by_label(
                                label=bond_label, orbitals=orb
                            )
                            orb_icohps_bndg.append(orb_icohp_bn)
                        bndg_orb_pair_list.append(orb)
                        bndg_orb_icohp_list.append(orb_icohps_bndg)
                        bndg_orb_integral_list.append(bndg_orb)
                        bndg_orb_perc_list.append(per_bndg_orb)

                    # replace nan values with zero (tackle numerical integration issues)
                    antibndg_orb = antibndg_orb if not np.isnan(antibndg_orb) else 0
                    per_anti_orb = per_anti_orb if not np.isnan(per_anti_orb) else 0
                    antibndg_tot = antibndg_tot if not np.isnan(antibndg_tot) else 0
                    per_anti_tot = per_anti_tot if not np.isnan(per_anti_tot) else 0
                    # skip collecting orb contributions if no summed antibonding contribution exists
                    if antibndg_tot > 0:
                        orb_icohps_anti = []
                        for bond_label in bond_labels:
                            orb_icohp_an = self.chemenv.Icohpcollection.get_icohp_by_label(
                                label=bond_label, orbitals=orb
                            )
                            orb_icohps_anti.append(orb_icohp_an)

                        antibndg_orb_pair_list.append(orb)
                        antibndg_orb_icohp_list.append(orb_icohps_anti)
                        antibndg_orb_integral_list.append(antibndg_orb)
                        antibndg_orb_perc_list.append(per_anti_orb)

                # Populate the dictionary with relevant orbitals for bonding interactions
                for inx, bndg_orb_pair in enumerate(bndg_orb_pair_list):
                    bndg_contri_perc = round(bndg_orb_integral_list[inx] / sum(bndg_orb_integral_list), 2)
                    # filter out very small bonding interactions (<self.orbital_cutoff)
                    if bndg_contri_perc > self.orbital_cutoff:
                        if bndg_orb_pair in orb_bonding_dict_data:
                            orb_bonding_dict_data[bndg_orb_pair].update(
                                {
                                    "orb_contribution_perc_bonding": bndg_contri_perc,
                                    "bonding": {
                                        "integral": bndg_orb_integral_list[inx],
                                        "perc": bndg_orb_perc_list[inx],
                                    },
                                }
                            )
                        else:
                            orb_bonding_dict_data[bndg_orb_pair] = {
                                f"I{type_pop}_mean": round(np.mean(bndg_orb_icohp_list[inx]), 4),
                                f"I{type_pop}_sum": round(np.sum(bndg_orb_icohp_list[inx]), 4),
                                "orb_contribution_perc_bonding": round(
                                    bndg_orb_integral_list[inx] / sum(bndg_orb_integral_list),
                                    2,
                                ),
                                "bonding": {
                                    "integral": bndg_orb_integral_list[inx],
                                    "perc": bndg_orb_perc_list[inx],
                                },
                            }

                # Populate the dictionary with relevant orbitals for antibonding interactions
                for inx, antibndg_orb_pair in enumerate(antibndg_orb_pair_list):
                    antibndg_contri_perc = round(
                        antibndg_orb_integral_list[inx] / sum(antibndg_orb_integral_list),
                        2,
                    )
                    # filter out very small antibonding interactions (<self.orbital_cutoff)
                    if antibndg_contri_perc > self.orbital_cutoff:
                        if antibndg_orb_pair in orb_bonding_dict_data:
                            orb_bonding_dict_data[antibndg_orb_pair].update(
                                {
                                    "orb_contribution_perc_antibonding": round(
                                        antibndg_orb_integral_list[inx] / sum(antibndg_orb_integral_list),
                                        2,
                                    ),
                                    "antibonding": {
                                        "integral": antibndg_orb_integral_list[inx],
                                        "perc": antibndg_orb_perc_list[inx],
                                    },
                                }
                            )
                        else:
                            orb_bonding_dict_data[antibndg_orb_pair] = {
                                f"I{type_pop}_mean": round(np.mean(antibndg_orb_icohp_list[inx]), 4),
                                f"I{type_pop}_sum": round(np.sum(antibndg_orb_icohp_list[inx]), 4),
                                "orb_contribution_perc_antibonding": round(
                                    antibndg_orb_integral_list[inx] / sum(antibndg_orb_integral_list),
                                    2,
                                ),
                                "antibonding": {
                                    "integral": antibndg_orb_integral_list[inx],
                                    "perc": antibndg_orb_perc_list[inx],
                                },
                            }

                orb_bonding_dict_data["relevant_bonds"] = bond_labels  # type: ignore

                orb_resolved_bond_info[bond_resolved_label_key] = orb_bonding_dict_data

        return orb_resolved_bond_info

    def _get_bond_resolved_data_stats(self, orb_resolved_bond_data: dict):
        """
        Retrieve the maximum bonding and anti-bonding orbital contributions.

        :param orb_resolved_bond_data: A dictionary with orbital names as keys and corresponding bonding data

        Returns:
            dict with orbital data stats the site for relevant orbitals, e.g.
            {'orbital_summary_stats': {'max_bonding_contribution': {'2s-3s': 0.68},
            'max_antibonding_contribution': {'2s-2pz': 0.36}}}

        """
        # get max orbital bonding and contribution for the site
        orb_pairs_bndg = []
        orb_pairs_antibndg = []
        orb_contri_bndg = []
        orb_contri_antibndg = []
        orbital_summary_stats = {"orbital_summary_stats": {}}  # type: ignore
        if orb_resolved_bond_data:
            for orb_pair, data in orb_resolved_bond_data.items():
                if "orb_contribution_perc_bonding" in data:
                    orb_pairs_bndg.append(orb_pair)
                    orb_contri_bndg.append(data["orb_contribution_perc_bonding"])

                if "orb_contribution_perc_antibonding" in data:
                    orb_pairs_antibndg.append(orb_pair)
                    orb_contri_antibndg.append(data["orb_contribution_perc_antibonding"])

            if orb_contri_bndg:
                max_orb_contri_bndg = max(orb_contri_bndg)
                max_orb_contri_bndg_inxs = [
                    inx for inx, orb_contri in enumerate(orb_contri_bndg) if orb_contri == max_orb_contri_bndg
                ]
                max_orb_contri_bndg_dict = {}
                for inx in max_orb_contri_bndg_inxs:
                    max_orb_contri_bndg_dict[orb_pairs_bndg[inx]] = orb_contri_bndg[inx]
                orbital_summary_stats["orbital_summary_stats"]["max_bonding_contribution"] = max_orb_contri_bndg_dict
            if orb_contri_antibndg:
                max_orb_contri_antibndg = max(orb_contri_antibndg)
                max_antibndg_contri_inxs = [
                    inx
                    for inx, orb_anti_per in enumerate(orb_contri_antibndg)
                    if orb_anti_per == max_orb_contri_antibndg
                ]
                max_antibndg_contri_dict = {}
                for inx in max_antibndg_contri_inxs:
                    max_antibndg_contri_dict[orb_pairs_antibndg[inx]] = orb_contri_antibndg[inx]
                orbital_summary_stats["orbital_summary_stats"]["max_antibonding_contribution"] = (
                    max_antibndg_contri_dict
                )

        return orbital_summary_stats

    def get_site_orbital_resolved_labels(self):
        """
        Return relevant orbitals and bond labels for each symmetrically independent site.

        Returns:
            dict with bond labels for each site for relevant orbitals, e.g.
            {'Na1: Na-Cl': {'3s-3s': ['21', '23', '24', '27', '28', '30']}

        """
        site_bond_labels = self.get_site_bond_resolved_labels()
        orb_plot_data = {atom_pair: {} for atom_pair in site_bond_labels}
        if self.orbital_resolved:
            for site_index, cba_data in self.condensed_bonding_analysis["sites"].items():
                for atom in cba_data["bonds"]:
                    for orb_pair in cba_data["bonds"][atom]["orbital_data"]:
                        if orb_pair not in ("orbital_summary_stats", "relevant_bonds"):
                            atom_pair = [cba_data["ion"], atom]
                            atom_pair.sort()
                            key = (
                                self.structure.sites[site_index].species_string
                                + str(site_index + 1)
                                + ": "
                                + "-".join(atom_pair)
                            )
                            label_list = site_bond_labels[key]
                            orb_plot_data[key].update({orb_pair: label_list})
        else:
            print("Please set orbital_resolved to True when instantiating Analysis object, to get this data")

        return orb_plot_data

    @staticmethod
    def _get_strenghts_for_each_bond(pairs: list[list[str]], strengths: list[float], nameion: str | None = None):
        """
        Return a dictionary of bond strengths.

        :param pairs: list of list including labels for the atoms, e.g., [['O3', 'Cu1'], ['O3', 'Cu1']]
        :param strengths: list that gives the icohp strengths as a float, [-1.86287, -1.86288]
        :param nameion: string including the name of the cation in the list, e.g Cu1

        Returns:
            dict including inormation on icohps for each bond type, e.g.
            {'Yb-Sb': [-1.59769, -2.14723, -1.7925, -1.60773, -1.80149, -2.14335]}


        """
        dict_strenghts = {}  # type: ignore

        for pair, strength in zip(pairs, strengths):
            if nameion is not None:
                new = [
                    LobsterNeighbors._split_string(pair[0])[0],
                    LobsterNeighbors._split_string(pair[1])[0],
                ]
                new = Analysis._sort_name(new, nameion)
                string_here = new[0] + "-" + new[1]
            else:
                new = sorted(
                    [
                        LobsterNeighbors._split_string(pair[0])[0],
                        LobsterNeighbors._split_string(pair[1])[0],
                    ]
                )
                string_here = new[0] + "-" + new[1]

            if string_here not in dict_strenghts:
                dict_strenghts[string_here] = []
            dict_strenghts[string_here].append(strength)
        return dict_strenghts

    @staticmethod
    def _sort_name(pair: list[str], nameion: str | None = None):
        """
        Place the cation first in a list of name strings.

        :param pair: ["O","Cu"]
        :param nameion: "Cu"

        Returns:
            will return list of str, e.g. ["Cu", "O"]

        """
        if nameion is not None:
            new = []
            if pair[0] == nameion:
                new.append(pair[0])
                new.append(pair[1])

            elif pair[1] == nameion:
                new.append(pair[1])
                new.append(pair[0])

        return new

    @staticmethod
    def _sort_orbital_atom_pair(
        atom_pair: list[str],
        label: str,
        complete_cohp: CompleteCohp,
        orb_pair: str,
    ):
        """
        Place the cation first in a list of name strings and add the associated orbital name alongside the atom name.

        :param atom_pair: list of atom pair with cation first eg., ["Cl","Na"]
        :param label: LOBSTER relevant bond label eg ., "3"
        :param complete_cohp: pymatgen CompleteCohp object
        :param orb_pair: relevant orbital pair eg., "2px-3s"

        Returns:
            will return list of str, e.g. ["Na(2px)", "Cl(3s)"]

        """
        orb_atom = {}  # type: ignore
        orb_pair_list = orb_pair.split("-")
        # get orbital associated to the atom and store in a dict
        for _inx, (site, site_orb) in enumerate(zip(complete_cohp.bonds[label]["sites"], orb_pair_list)):
            if site.species_string in orb_atom:  # check necessary for bonds between same atoms
                orb_atom[site.species_string].append(site_orb)
            else:
                orb_atom[site.species_string] = [site_orb]

        orb_atom_list = []
        # add orbital name next to atom_pair
        for inx, atom in enumerate(atom_pair):
            # check to ensure getting 2nd orbital if bond is between same atomic species
            if inx == 1 and len(orb_atom.get(atom)) > 1:  # type: ignore
                atom_with_orb_name = f"{atom}({orb_atom.get(atom)[1]})"  # type: ignore
            else:
                atom_with_orb_name = f"{atom}({orb_atom.get(atom)[0]})"  # type: ignore
            orb_atom_list.append(atom_with_orb_name)

        return orb_atom_list

    def _get_antibdg_states(self, cohps, labels: list[str], nameion: str | None = None, limit=0.01):
        """
        Return a dictionary containing information on anti-bonding states.

        e.g., similar to: {'Cu-O': True, 'Cu-F': True}

        :param cohps: list of pymatgen.electronic_structure.cohp.Cohp objects
        :param labels: ['2 x Cu-O', '4 x Cu-F']
        :param nameion: string of the cation name, e.g. "Cu"
        :param limit: limit to detect antibonding states

        Returns:
            dict including in formation on whether antibonding interactions exist,
            e.g., {'Cu-O': True, 'Cu-F': True}


        """
        dict_antibd = {}
        for label, cohp in zip(labels, cohps):
            if label is not None:
                if nameion is not None:
                    new = label.split(" ")[2].split("-")
                    sorted_new = self._sort_name(new, nameion)
                    new_label = sorted_new[0] + "-" + sorted_new[1]
                else:
                    new = label.split(" ")[2].split("-")
                    sorted_new = sorted(new.copy())
                    new_label = sorted_new[0] + "-" + sorted_new[1]

                antbd = cohp.has_antibnd_states_below_efermi(limit=limit)
                if Spin.down in antbd:
                    dict_antibd[new_label] = antbd[Spin.up] or antbd[Spin.down]
                else:
                    dict_antibd[new_label] = antbd[Spin.up]

        return dict_antibd

    def _integrate_antbdstates_below_efermi_for_set_cohps(self, labels: list[str], cohps, nameion: str):
        """
        Return a dictionary containing information on antibonding states.

        .. warning:: NEEDS MORE TESTS

        It is important to note that only the energy range that has been computed can be considered
        (i.e., this might not be all)

        e.g. output: {'Cu-O': {'integral': 4.24374775705, 'perc': 5.7437713186999995},
        'Cu-F': {'integral': 3.07098300965, 'perc': 4.25800841445}}

        :param cohps: list of pymatgen.electronic_structure.cohp.Cohp objects
        :param labels: ['2 x Cu-O', '4 x Cu-F']
        :param nameion: string of the cation name, e.g. "Cu"

        Returns:
            dict including in formation on whether antibonding interactions exist,
            e.g., {'Cu-O': {'integral': 4.24374775705, 'perc': 5.7437713186999995},
            'Cu-F': {'integral': 3.07098300965, 'perc': 4.25800841445}}}
        """
        dict_bd_antibd = {}
        for label, cohp in zip(labels, cohps):
            if label is not None:
                new = label.split(" ")[2].split("-")
                sorted_new = self._sort_name(new, nameion)
                new_label = sorted_new[0] + "-" + sorted_new[1]
                if not self.are_cobis and not self.are_coops:
                    (
                        integral,
                        perc,
                        integral2,
                        perc2,
                    ) = self._integrate_antbdstates_below_efermi(cohp, start=self.start)
                else:
                    (
                        integral2,
                        perc2,
                        integral,
                        perc,
                    ) = self._integrate_antbdstates_below_efermi(cohp, start=self.start)

                if integral == 0 and integral2 != 0.0:
                    dict_bd_antibd[new_label] = {
                        "bonding": {"integral": integral2, "perc": perc2},
                        "antibonding": {"integral": integral, "perc": 0.0},
                    }
                elif integral2 == 0.0 and integral != 0.0:
                    dict_bd_antibd[new_label] = {
                        "bonding": {"integral": integral2, "perc": 0.0},
                        "antibonding": {"integral": integral, "perc": perc},
                    }
                elif integral == 0.0 and integral2 == 0.0:
                    dict_bd_antibd[new_label] = {
                        "bonding": {"integral": integral2, "perc": 0.0},
                        "antibonding": {"integral": integral, "perc": 0.0},
                    }
                else:
                    dict_bd_antibd[new_label] = {
                        "bonding": {"integral": integral2, "perc": perc2},
                        "antibonding": {"integral": integral, "perc": perc},
                    }

        return dict_bd_antibd

    def _integrate_antbdstates_below_efermi(self, cohp, start: float | None):
        """
        Integrate the cohp data to compute bonding and anti-bonding contribution below efermi.

        .. warning:: NEEDS MORE TESTS

        This integrates the whole COHP curve that has been computed.
        The energy range is very important.
        At present the energy range considered is dependent on COHPstartEnergy
        set during lobster runs. The bonding / antibonding integral values are sensitive to this parameter.
        If COHPstartEnergy value does not cover entire range of VASP calculations then
        absolute value of ICOHP_sum might not be equivalent to (bonding- antibonding) integral values.

        :param cohp: cohp object
        :param start: integration start energy in eV , eg start = -15

        Returns:
            absolute value of antibonding, percentage value of antibonding,
            absolute value of bonding and percentage value of bonding interactions
        """
        warnings.warn(
            "The bonding, antibonding integral/percent values are numerical estimate."
            " These values are sensitive to COHPstartEnergy parameter."
            " If COHPstartEnergy value does not cover entire range of VASP calculations then"
            " absolute value of ICOHP_sum might not be equivalent to (bonding- antibonding) integral values."
        )

        from scipy.integrate import trapezoid

        def integrate_positive(y, x):
            """
            Integrate only bonding interactions of COHPs.

            :param y: COHP values
            :param x: Energy values

            Returns:
                integrated value of bonding interactions
            """
            y = np.asanyarray(y)
            x = np.asanyarray(x)

            bonding = trapezoid(y, x)

            return np.round(bonding, 2)

        def integrate_negative(y, x):
            """
            Integrate only anti-bonding interactions of COHPs.

            :param y: COHP values
            :param x: Energy values

            Returns:
                integrated value of anti-bonding interactions
            """
            y = np.asanyarray(y)
            x = np.asanyarray(x)
            antibonding = trapezoid(y, x)

            return np.round(antibonding, 2)

        # will integrate spin.up and spin.down only below efermi
        energies_corrected = cohp.energies - cohp.efermi
        summedcohp = cohp.cohp[Spin.up] + cohp.cohp[Spin.down] if Spin.down in cohp.cohp else cohp.cohp[Spin.up]

        cohp_bf = []
        en_bf = []

        for i, en in enumerate(energies_corrected):
            if (start is None) and en <= 0:
                en_bf.append(en)
                cohp_bf.append(-1 * summedcohp[i])
            if (start is not None) and 0 >= en >= start:
                en_bf.append(en)
                cohp_bf.append(-1 * summedcohp[i])

        # Separate the bonding and antibonding COHP values in separate lists
        pos = []
        en_pos = []
        neg = []
        en_neg = []

        for i, scohp in enumerate(cohp_bf):
            if scohp >= 0:
                pos.append(scohp)
                en_pos.append(energies_corrected[i])

        for i, scohp in enumerate(cohp_bf):
            if scohp <= 0:
                neg.append(-1 * scohp)
                en_neg.append(energies_corrected[i])

        antibonding = integrate_negative(y=neg, x=en_neg)

        bonding = integrate_positive(y=pos, x=en_pos)

        return (
            antibonding,
            np.round(abs(antibonding) / (abs(bonding) + abs(antibonding)), 5),
            bonding,
            np.round(abs(bonding) / (abs(bonding) + abs(antibonding)), 5),
        )

    def _get_pop_type(self):
        """
        Return the type of the input population file.

        Returns:
            A String of analysed population can be COOP/COBI/COHP
        """
        if self.are_cobis:
            type_pop = "COBI"
        elif self.are_coops:
            type_pop = "COOP"
        else:
            type_pop = "COHP"

        return type_pop

    @staticmethod
    def _get_bond_dict(
        bond_strength_dict: dict,
        small_antbd_dict: dict,
        nameion: str | None = None,
        large_antbd_dict: dict | None = None,
        type_pop: str | None = None,
    ):
        """
        Return a bond_dict that contains information for each site.

        :param bond_strength_dict: dict with bond names as key and lists of bond strengths as items
        :param small_antbd_dict: dict including if there are antibonding interactions, {'Yb-Sb': False}
        :param nameion: name of the cation, e.g. Yb
        :param large_antbd_dict: will be implemented later
        :param type_pop: population type analyzed. eg. COHP

        Returns:
            Eg., if type_pop == 'COHP', will return
            dict including information on the anion (as label) and the ICOHPs in the item of the dict
            ICOHP_mean refers to the mean ICOHP in eV
            ICOHP_sum refers to the sum of the ICOHPs in eV
            has_antibdg_states_below_Efermi is True if there are antibonding interactions below Efermi
            "number_of_bonds" will count the numbers of bonds to the cation

        Example:
            {'Sb': {'ICOHP_mean': '-1.85', 'ICOHP_sum': '-11.09',
            'has_antibdg_states_below_Efermi': False, 'number_of_bonds': 6}}

        """
        bond_dict = {}

        for key, item in bond_strength_dict.items():
            if nameion is not None:
                a = key.split("-")[0]
                b = key.split("-")[1]
                if a == nameion:
                    key_here = b
                elif b == nameion:
                    key_here = a

            if large_antbd_dict is None:
                bond_dict[key_here] = {
                    f"I{type_pop}_mean": str(round(np.mean(item), 2)),
                    f"I{type_pop}_sum": str(round(np.sum(item), 2)),
                    "has_antibdg_states_below_Efermi": small_antbd_dict[key],
                    "number_of_bonds": len(item),
                }
            else:
                bond_dict[key_here] = {
                    f"I{type_pop}_mean": str(round(np.mean(item), 2)),
                    f"I{type_pop}_sum": str(round(np.sum(item), 2)),
                    "has_antibdg_states_below_Efermi": small_antbd_dict[key],
                    "number_of_bonds": len(item),
                    "perc_antibdg_states_below_Efermi": large_antbd_dict[key],
                }

        return bond_dict

    def set_condensed_bonding_analysis(self):
        """
        Condense the bonding analysis into a summary dictionary.

        Returns:
            None

        """
        self.condensed_bonding_analysis = {}
        # which icohps are considered
        if self.which_bonds == "cation-anion":
            limit_icohps = self.chemenv._get_limit_from_extremum(
                self.chemenv.Icohpcollection,
                self.cutoff_icohp,
                adapt_extremum_to_add_cond=True,
                additional_condition=1,
            )
        elif self.which_bonds == "all":
            limit_icohps = self.chemenv._get_limit_from_extremum(
                self.chemenv.Icohpcollection,
                self.cutoff_icohp,
                adapt_extremum_to_add_cond=True,
                additional_condition=0,
            )
            # formula of the compound
        formula = str(self.structure.composition.reduced_formula)
        # set population type
        type_pop = self._get_pop_type()
        # how many inequivalent cations are in the structure
        if self.which_bonds == "cation-anion":
            number_considered_ions = len(self.seq_ineq_ions)
        elif self.which_bonds == "all":
            number_considered_ions = len(self.seq_ineq_ions)

        # what was the maximum bond lengths that was considered
        max_bond_lengths = max(self.chemenv.Icohpcollection._list_length)

        # what are the charges for the cations in the structure
        charge_list = self.chemenv.valences

        # dictionary including bonding information for each site
        site_dict = {}
        if self.which_bonds == "cation-anion":
            for ication, ce, cation_anion_infos, labels, cohps in zip(
                self.seq_ineq_ions,
                self.seq_coord_ions,
                self.seq_infos_bonds,
                self.seq_labels_cohps,
                self.seq_cohps,
            ):
                namecation = str(self.structure[ication].specie)

                # This will compute the mean strengths of ICOHPs
                mean_icohps = self._get_strenghts_for_each_bond(
                    pairs=cation_anion_infos[4],
                    strengths=cation_anion_infos[1],
                    nameion=namecation,
                )
                # pairs, strengths, nameion
                # will collect if there are antibonding states present
                antbdg = self._get_antibdg_states(cohps, labels, namecation)
                dict_antibonding = self._integrate_antbdstates_below_efermi_for_set_cohps(
                    labels, cohps, nameion=namecation
                )
                bond_dict = self._get_bond_dict(mean_icohps, antbdg, namecation, type_pop=type_pop)
                bond_resolved_labels = self.get_site_bond_resolved_labels()

                for cation_name, icohp_data in bond_dict.items():
                    for atom_pair, bonding_data in dict_antibonding.items():
                        if namecation == atom_pair.split("-")[0] and cation_name == atom_pair.split("-")[1]:
                            icohp_data["bonding"] = bonding_data["bonding"]
                            icohp_data["antibonding"] = bonding_data["antibonding"]
                            if self.orbital_resolved:
                                # get orb resolved data to be added
                                orb_resolved_bond_info = self._get_orbital_resolved_data(
                                    nameion=namecation,
                                    iion=ication,
                                    labels=labels,
                                    bond_resolved_labels=bond_resolved_labels,
                                    type_pop=type_pop,
                                )
                                # match the dict key in bond_dict and get corresponding orbital data
                                for ion_atom_pair_orb in orb_resolved_bond_info:
                                    orb_data_atom_pair = ion_atom_pair_orb.split(": ")[-1]
                                    atom_pair_here = atom_pair.split("-")
                                    atom_pair_here.sort()
                                    if (
                                        orb_data_atom_pair == "-".join(atom_pair_here)
                                        and (namecation + str(ication + 1) + ":") in ion_atom_pair_orb
                                    ):
                                        icohp_data["orbital_data"] = orb_resolved_bond_info[ion_atom_pair_orb]

                                        orb_data_stats = self._get_bond_resolved_data_stats(
                                            orb_resolved_bond_data=orb_resolved_bond_info[ion_atom_pair_orb],
                                        )

                                        icohp_data["orbital_data"].update(orb_data_stats)

                site_dict[ication] = {
                    "env": ce,
                    "bonds": bond_dict,
                    "ion": namecation,
                    "charge": charge_list[ication],
                    "relevant_bonds": cation_anion_infos[3],
                }
        elif self.which_bonds == "all":
            for iion, ce, bond_infos, labels, cohps in zip(
                self.seq_ineq_ions,
                self.seq_coord_ions,
                self.seq_infos_bonds,
                self.seq_labels_cohps,
                self.seq_cohps,
            ):
                nameion = str(self.structure[iion].specie)

                # This will compute the mean strengths of ICOHPs
                mean_icohps = self._get_strenghts_for_each_bond(
                    pairs=bond_infos[4], strengths=bond_infos[1], nameion=None
                )
                # pairs, strengths, nameion
                # will collect if there are antibonding states present
                antbdg = self._get_antibdg_states(cohps, labels, nameion=None)

                dict_antibonding = self._integrate_antbdstates_below_efermi_for_set_cohps(labels, cohps, nameion)

                bond_dict = self._get_bond_dict(mean_icohps, antbdg, nameion=nameion, type_pop=type_pop)
                bond_resolved_labels = self.get_site_bond_resolved_labels()

                for cation_name, icohp_data in bond_dict.items():
                    for atom_pair, bonding_data in dict_antibonding.items():
                        if nameion == atom_pair.split("-")[0] and cation_name == atom_pair.split("-")[1]:
                            icohp_data["bonding"] = bonding_data["bonding"]
                            icohp_data["antibonding"] = bonding_data["antibonding"]
                            if self.orbital_resolved:
                                # get orb resolved data to be added
                                orb_resolved_bond_info = self._get_orbital_resolved_data(
                                    nameion=nameion,
                                    iion=iion,
                                    labels=labels,
                                    bond_resolved_labels=bond_resolved_labels,
                                    type_pop=type_pop,
                                )
                                # match the dict key in bond_dict and get corresponding orbital data
                                for ion_atom_pair_orb in orb_resolved_bond_info:
                                    orb_data_atom_pair = ion_atom_pair_orb.split(": ")[-1]
                                    atom_pair_here = atom_pair.split("-")
                                    atom_pair_here.sort()
                                    if (
                                        orb_data_atom_pair == "-".join(atom_pair_here)
                                        and (nameion + str(iion + 1) + ":") in ion_atom_pair_orb
                                    ):
                                        icohp_data["orbital_data"] = orb_resolved_bond_info[ion_atom_pair_orb]

                                        orb_data_stats = self._get_bond_resolved_data_stats(
                                            orb_resolved_bond_data=orb_resolved_bond_info[ion_atom_pair_orb],
                                        )

                                        icohp_data["orbital_data"].update(orb_data_stats)

                site_dict[iion] = {
                    "env": ce,
                    "bonds": bond_dict,
                    "ion": nameion,
                    "charge": charge_list[iion],
                    "relevant_bonds": bond_infos[3],
                }

        if self.path_to_madelung is None and self.madelung_obj is None:
            if self.which_bonds == "cation-anion":
                # This sets the dictionary including the most important information on the compound
                self.condensed_bonding_analysis = {
                    "formula": formula,
                    "max_considered_bond_length": max_bond_lengths,
                    f"limit_i{type_pop.lower()}": limit_icohps,
                    "number_of_considered_ions": number_considered_ions,
                    "sites": site_dict,
                    "type_charges": self.type_charge,
                }
            elif self.which_bonds == "all":
                self.condensed_bonding_analysis = {
                    "formula": formula,
                    "max_considered_bond_length": max_bond_lengths,
                    f"limit_i{type_pop.lower()}": limit_icohps,
                    "number_of_considered_ions": number_considered_ions,
                    "sites": site_dict,
                    "type_charges": self.type_charge,
                }
        else:
            madelung = MadelungEnergies(self.path_to_madelung) if self.path_to_madelung else self.madelung_obj
            if self.type_charge == "Mulliken":
                madelung_energy = madelung.madelungenergies_mulliken
            elif self.type_charge == "Löwdin":
                madelung_energy = madelung.madelungenergies_loewdin
            else:
                madelung_energy = None
            # This sets the dictionary including the most important information on the compound
            if self.which_bonds == "cation-anion":
                self.condensed_bonding_analysis = {
                    "formula": formula,
                    "max_considered_bond_length": max_bond_lengths,
                    f"limit_i{type_pop.lower()}": limit_icohps,
                    "number_of_considered_ions": number_considered_ions,
                    "sites": site_dict,
                    "type_charges": self.type_charge,
                    "madelung_energy": madelung_energy,
                }
            elif self.which_bonds == "all":
                self.condensed_bonding_analysis = {
                    "formula": formula,
                    "max_considered_bond_length": max_bond_lengths,
                    f"limit_i{type_pop.lower()}": limit_icohps,
                    "number_of_considered_ions": number_considered_ions,
                    "sites": site_dict,
                    "type_charges": self.type_charge,
                    "madelung_energy": madelung_energy,
                }

    def set_summary_dicts(self):
        """
        Set summary dict that can be used for correlations.

        bond_dict that includes information on each bond

        "has_antbd" tells if there are antbonding states
        "ICOHP_mean" shows the mean of all ICOHPs in EV

        {'Yb-Sb': { 'has_antbdg': False, 'ICOHP_mean': -1.7448},
        'Mn-Sb': { 'has_antbdg': True, 'ICOHP_mean': -1.525}}

        a cation dict that includes all different coordination environments and counts for them
        {'Na': {'T:4': 4, 'A:2': 4}, 'Si': {'T:6': 4, 'PP:6': 4}}

        Returns:
            None

        """
        relevant_ion_ids = [isite for isite in self.list_equivalent_sites if isite in self.seq_ineq_ions]
        # set population type
        type_pop = self._get_pop_type()

        final_dict_bonds = {}
        for key in relevant_ion_ids:
            item = self.condensed_bonding_analysis["sites"][key]
            for type, properties in item["bonds"].items():
                label_list = [item["ion"], str(type)]
                new_label = sorted(label_list.copy())
                label = str(new_label[0]) + "-" + str(new_label[1])

                if label not in final_dict_bonds:
                    final_dict_bonds[label] = {
                        "number_of_bonds": int(properties["number_of_bonds"]),
                        f"I{type_pop}_sum": float(properties[f"I{type_pop}_sum"]),
                        "has_antbdg": properties["has_antibdg_states_below_Efermi"],
                    }
                else:
                    final_dict_bonds[label]["number_of_bonds"] += int(properties["number_of_bonds"])
                    final_dict_bonds[label][f"I{type_pop}_sum"] += float(properties[f"I{type_pop}_sum"])
                    final_dict_bonds[label]["has_antbdg"] = (
                        final_dict_bonds[label]["has_antbdg"] or properties["has_antibdg_states_below_Efermi"]
                    )
        self.final_dict_bonds = {}
        for key, item in final_dict_bonds.items():
            self.final_dict_bonds[key] = {}
            self.final_dict_bonds[key][f"I{type_pop}_mean"] = item[f"I{type_pop}_sum"] / (item["number_of_bonds"])
            self.final_dict_bonds[key]["has_antbdg"] = item["has_antbdg"]

        # rework, add all environments!
        final_dict_ions = {}
        for key in relevant_ion_ids:
            if self.condensed_bonding_analysis["sites"][key]["ion"] not in final_dict_ions:
                final_dict_ions[self.condensed_bonding_analysis["sites"][key]["ion"]] = [
                    self.condensed_bonding_analysis["sites"][key]["env"]
                ]
            else:
                final_dict_ions[self.condensed_bonding_analysis["sites"][key]["ion"]].append(
                    self.condensed_bonding_analysis["sites"][key]["env"]
                )

        self.final_dict_ions = {}
        for key, item in final_dict_ions.items():
            self.final_dict_ions[key] = dict(Counter(item))

    @staticmethod
    def get_lobster_calc_quality_summary(
        path_to_poscar: str | None = None,
        path_to_lobsterout: str | None = None,
        path_to_lobsterin: str | None = None,
        path_to_potcar: str | None = None,
        potcar_symbols: list | None = None,
        path_to_charge: str | None = None,
        path_to_bandoverlaps: str | None = None,
        path_to_doscar: str | None = None,
        path_to_vasprun: str | None = None,
        structure_obj: Structure | None = None,
        lobsterin_obj: Lobsterin | None = None,
        lobsterout_obj: Lobsterout | None = None,
        charge_obj: Charge | None = None,
        bandoverlaps_obj: Bandoverlaps | None = None,
        lobster_completedos_obj: LobsterCompleteDos | None = None,
        vasprun_obj: Vasprun | None = None,
        dos_comparison: bool = False,
        e_range: list = [-5, 0],
        n_bins: int | None = None,
        bva_comp: bool = False,
    ) -> dict:
        """
        Analyze LOBSTER calculation quality.

        :param path_to_poscar: path to structure file
        :param path_to_lobsterout: path to lobsterout file
        :param path_to_lobsterin: path to lobsterin file
        :param path_to_potcar: path to VASP potcar file
        :param potcar_symbols: list of potcar symbols from potcar file (can be used if no potcar available)
        :param path_to_charge: path to CHARGE.lobster file
        :param path_to_bandoverlaps: path to bandOverlaps.lobster file
        :param path_to_doscar: path to DOSCAR.lobster or DOSCAR.LSO.lobster file
        :param path_to_vasprun: path to vasprun.xml file
        :param structure_obj: pymatgen pymatgen.core.structure.Structure object
        :param lobsterin_obj: pymatgen.lobster.io.Lobsterin object
        :param lobsterout_obj: pymatgen lobster.io.Lobsterout object
        :param charge_obj: pymatgen lobster.io.Charge object
        :param bandoverlaps_obj: pymatgen lobster.io.BandOverlaps object
        :param lobster_completedos_obj: pymatgen.electronic_structure.dos.LobsterCompleteDos object
        :param vasprun_obj: pymatgen vasp.io.Vasprun object
        :param dos_comparison: will compare DOS from VASP and LOBSTER and return tanimoto index
        :param e_range: energy range for DOS comparisons
        :param n_bins: number of bins to discretize DOS for comparisons
        :param bva_comp: Compares LOBSTER charge signs with Bond valence charge signs

        Returns:
            A dict of summary of LOBSTER calculation quality by analyzing basis set used,
            charge spilling from lobsterout/ PDOS comparisons of VASP and LOBSTER /
            BVA charge comparisons

        """
        quality_dict = {}

        if path_to_potcar and not potcar_symbols and not path_to_vasprun and not vasprun_obj:
            potcar_names = Lobsterin._get_potcar_symbols(POTCAR_input=path_to_potcar)
        elif not path_to_potcar and not path_to_vasprun and not vasprun_obj and potcar_symbols:
            potcar_names = potcar_symbols
        elif path_to_vasprun and not vasprun_obj:
            vasprun = Vasprun(path_to_vasprun, parse_potcar_file=False, parse_eigen=False)
            potcar_names = [potcar.split(" ")[1] for potcar in vasprun.potcar_symbols]
        elif vasprun_obj and not path_to_vasprun:
            potcar_names = [potcar.split(" ")[1] for potcar in vasprun_obj.potcar_symbols]
        else:
            raise ValueError(
                "Please provide either path_to_potcar or list of "
                "potcar_symbols or path to vasprun.xml or vasprun object. "
                "Crucial to identify basis used for projections"
            )

        if path_to_poscar:
            struct = Structure.from_file(path_to_poscar)
        elif structure_obj:
            struct = structure_obj
        else:
            raise ValueError("Please provide path_to_poscar or structure_obj")

        ref_bases = Lobsterin.get_all_possible_basis_functions(structure=struct, potcar_symbols=potcar_names)

        if path_to_lobsterin:
            lobs_in = Lobsterin.from_file(path_to_lobsterin)
        elif lobsterin_obj:
            lobs_in = lobsterin_obj
        else:
            raise ValueError("Please provide path_to_lobsterin or lobsterin_obj")

        calc_basis = []
        for basis in lobs_in["basisfunctions"]:
            basis_sep = basis.split()[1:]
            basis_comb = " ".join(basis_sep)
            calc_basis.append(basis_comb)

        if calc_basis == list(ref_bases[0].values()):
            quality_dict["minimal_basis"] = True  # type: ignore
        else:
            quality_dict["minimal_basis"] = False  # type: ignore
            warnings.warn(
                "Consider rerunning the calc with the minimum basis as well. Choosing is "
                "larger basis set is recommended if you see a significant improvement of "
                "the charge spilling and material has non-zero band gap."
            )

        if path_to_lobsterout:
            lob_out = Lobsterout(path_to_lobsterout)
        elif lobsterout_obj:
            lob_out = lobsterout_obj
        else:
            raise ValueError("Please provide path_to_lobsterout or lobsterout_obj")

        quality_dict["charge_spilling"] = {
            "abs_charge_spilling": round((sum(lob_out.charge_spilling) / 2) * 100, 4),
            "abs_total_spilling": round((sum(lob_out.total_spilling) / 2) * 100, 4),
        }  # type: ignore

        if path_to_bandoverlaps is not None and not bandoverlaps_obj:
            band_overlaps = Bandoverlaps(filename=path_to_bandoverlaps) if Path(path_to_bandoverlaps).exists() else None
        elif path_to_bandoverlaps is None and bandoverlaps_obj:
            band_overlaps = bandoverlaps_obj
        else:
            band_overlaps = None

        if band_overlaps is not None:
            for line in lob_out.warning_lines:
                if "k-points could not be orthonormalized" in line:
                    total_kpoints = int(line.split(" ")[2])

            # store actual number of devations above pymatgen default limit of 0.1
            dev_val = []
            for dev in band_overlaps.max_deviation:
                if dev > 0.1:
                    dev_val.append(dev)

            quality_dict["band_overlaps_analysis"] = {  # type: ignore
                "file_exists": True,
                "limit_maxDeviation": 0.1,
                "has_good_quality_maxDeviation": band_overlaps.has_good_quality_maxDeviation(limit_maxDeviation=0.1),
                "max_deviation": round(max(band_overlaps.max_deviation), 4),
                "percent_kpoints_abv_limit": round((len(dev_val) / total_kpoints) * 100, 4),
            }

        else:
            quality_dict["band_overlaps_analysis"] = {  # type: ignore
                "file_exists": False,
                "limit_maxDeviation": None,
                "has_good_quality_maxDeviation": True,
                "max_deviation": None,
                "percent_kpoints_abv_limit": None,
            }

        if bva_comp:
            try:
                bond_valence = BVAnalyzer()

                bva_oxi = []
                if path_to_charge and not charge_obj:
                    lobs_charge = Charge(filename=path_to_charge)
                elif not path_to_charge and charge_obj:
                    lobs_charge = charge_obj
                else:
                    raise Exception("BVA comparison is requested, thus please provide path_to_charge or charge_obj")
                for i in bond_valence.get_valences(structure=struct):
                    if i >= 0:
                        bva_oxi.append("POS")
                    else:
                        bva_oxi.append("NEG")

                mull_oxi = []
                for i in lobs_charge.Mulliken:
                    if i >= 0:
                        mull_oxi.append("POS")
                    else:
                        mull_oxi.append("NEG")

                loew_oxi = []
                for i in lobs_charge.Loewdin:
                    if i >= 0:
                        loew_oxi.append("POS")
                    else:
                        loew_oxi.append("NEG")

                quality_dict["charge_comparisons"] = {}  # type: ignore
                if mull_oxi == bva_oxi:
                    quality_dict["charge_comparisons"]["bva_mulliken_agree"] = True  # type: ignore
                else:
                    quality_dict["charge_comparisons"]["bva_mulliken_agree"] = False  # type: ignore

                if mull_oxi == bva_oxi:
                    quality_dict["charge_comparisons"]["bva_loewdin_agree"] = True  # type: ignore
                else:
                    quality_dict["charge_comparisons"]["bva_loewdin_agree"] = False  # type: ignore

            except ValueError:
                quality_dict["charge_comparisons"] = {}  # type: ignore
                warnings.warn(
                    "Oxidation states from BVA analyzer cannot be determined. "
                    "Thus BVA charge comparison will be skipped"
                )
        if dos_comparison:
            if "LSO" not in str(path_to_doscar).split("."):
                warnings.warn(
                    "Consider using DOSCAR.LSO.lobster, as non LSO DOS from LOBSTER can have negative DOS values"
                )
            if path_to_doscar:
                doscar_lobster = Doscar(
                    doscar=path_to_doscar,
                    structure_file=path_to_poscar,
                    structure=structure_obj,
                )

                dos_lobster = doscar_lobster.completedos
            elif lobster_completedos_obj:
                dos_lobster = lobster_completedos_obj
            else:
                raise ValueError(
                    "Dos comparison is requested, so please provide either path_to_doscar or lobster_completedos_obj"
                )

            if path_to_vasprun:
                vasprun = Vasprun(path_to_vasprun, parse_potcar_file=False, parse_eigen=False)
            elif vasprun_obj:
                vasprun = vasprun_obj
            else:
                raise ValueError(
                    "Dos comparison is requested, so please provide either path to vasprun.xml or vasprun_obj"
                )
            dos_vasp = vasprun.complete_dos

            quality_dict["dos_comparisons"] = {}  # type: ignore

            for orb in dos_lobster.get_spd_dos():
                if e_range[0] >= min(dos_vasp.energies) and e_range[0] >= min(dos_lobster.energies):
                    min_e = e_range[0]
                else:
                    warnings.warn(
                        "Minimum energy range requested for DOS comparisons is not available "
                        "in VASP or LOBSTER calculation. Thus, setting min_e to -5 eV"
                    )
                    min_e = -5

                if e_range[-1] <= max(dos_vasp.energies) and e_range[-1] <= max(dos_lobster.energies):
                    max_e = e_range[-1]
                else:
                    warnings.warn(
                        "Maximum energy range requested for DOS comparisons is not available "
                        "in VASP or LOBSTER calculation. Thus, setting max_e to 0 eV"
                    )
                    max_e = 0

                if np.diff(dos_vasp.energies)[0] >= 0.1 or np.diff(dos_lobster.energies)[0] >= 0.1:
                    warnings.warn(
                        "Input DOS files have very few points in the energy interval and thus "
                        "comparisons will not be reliable. Please rerun the calculations with "
                        "higher number of DOS points. Set NEDOS and COHPSteps tags to >= 2000 in VASP and LOBSTER "
                        "calculations, respectively."
                    )

                if not n_bins:
                    n_bins = 56

                fp_lobster_orb = dos_lobster.get_dos_fp(
                    min_e=min_e,
                    max_e=max_e,
                    n_bins=n_bins,
                    normalize=True,
                    type=orb.name,
                )
                fp_vasp_orb = dos_vasp.get_dos_fp(
                    min_e=min_e,
                    max_e=max_e,
                    n_bins=n_bins,
                    normalize=True,
                    type=orb.name,
                )

                tani_orb = round(
                    dos_vasp.get_dos_fp_similarity(fp_lobster_orb, fp_vasp_orb, tanimoto=True),
                    4,
                )
                quality_dict["dos_comparisons"][f"tanimoto_orb_{orb.name}"] = tani_orb  # type: ignore

            fp_lobster = dos_lobster.get_dos_fp(
                min_e=min_e,
                max_e=max_e,
                n_bins=n_bins,
                normalize=True,
                type="summed_pdos",
            )
            fp_vasp = dos_vasp.get_dos_fp(
                min_e=min_e,
                max_e=max_e,
                n_bins=n_bins,
                normalize=True,
                type="summed_pdos",
            )

            tanimoto_summed = round(dos_vasp.get_dos_fp_similarity(fp_lobster, fp_vasp, tanimoto=True), 4)
            quality_dict["dos_comparisons"]["tanimoto_summed"] = tanimoto_summed  # type: ignore
            quality_dict["dos_comparisons"]["e_range"] = [min_e, max_e]  # type: ignore
            quality_dict["dos_comparisons"]["n_bins"] = n_bins  # type: ignore

        return quality_dict



# Copyright (c) lobsterpy development team
# Distributed under the terms of a BSD 3-Clause "New" or "Revised" License

"""This module defines classes to describe the COHPs automatically."""

from __future__ import annotations

import warnings
from pathlib import Path

from lobsterpy.plotting import InteractiveCohpPlotter, PlainCohpPlotter


class Description:
    """
    Base class that will write generate a text description for all relevant bonds.

    It analyses all relevant coordination environments in the system based on electronic structure theory.

    """

    def __init__(self, analysis_object):
        """
        Generate a text description for all relevant bonds.

        :param analysis_object: Analysis object from lobsterpy.analysis
        """
        self.analysis_object = analysis_object
        self.set_description()

    def set_description(self):
        """
        Set the descriptions of the structures using the cation names, starting with numbers at 1.

        Uses the cation names from the lobster files.

        Returns:
            None

        """
        self.condensed_bonding_analysis = self.analysis_object.condensed_bonding_analysis
        # set type of population analyzed
        type_pop = self.analysis_object._get_pop_type()
        # set units for populations
        units = " eV" if type_pop == "COHP" else ""
        if self.analysis_object.which_bonds == "cation-anion":
            relevant_cations = ", ".join(
                [
                    str(site.specie) + str(isite + 1)
                    for isite, site in enumerate(self.analysis_object.structure)
                    if isite in self.analysis_object.seq_ineq_ions
                ]
            )
            self.text = []
            self.text.append(
                "The compound "
                + str(self.condensed_bonding_analysis["formula"])
                + " has "
                + str(self.condensed_bonding_analysis["number_of_considered_ions"])
                + " symmetry-independent cation(s) with relevant cation-anion interactions: "
                + str(relevant_cations)
                + "."
            )

            for key, item in self.condensed_bonding_analysis["sites"].items():
                # It has 3 Ta-N (mean ICOHP: -4.78 eV, antibonding interactions below EFermi),
                bond_info = []
                orb_info = []
                for type, properties in item["bonds"].items():
                    if not properties["has_antibdg_states_below_Efermi"]:
                        bond_info.append(
                            str(properties["number_of_bonds"])
                            + " "
                            + item["ion"]
                            + "-"
                            + str(type)
                            + f" (mean I{type_pop}: "
                            ""
                            + properties[f"I{type_pop}_mean"]
                            + f"{units}, 0.0 percent antibonding interaction below EFermi)"
                        )
                        if self.analysis_object.orbital_resolved:
                            text_orbital = self._generate_orbital_resolved_analysis_text(
                                orbital_resolved_data=properties,
                                type_pop=type_pop,
                                atom_name=str(type),
                                ion=item["ion"],
                            )
                            orb_info.extend(text_orbital)
                    else:
                        bond_info.append(
                            str(properties["number_of_bonds"])
                            + " "
                            + item["ion"]
                            + "-"
                            + str(type)
                            + f" (mean I{type_pop}: "
                            ""
                            + properties[f"I{type_pop}_mean"]
                            + f"{units}, "
                            + str(round(properties["antibonding"]["perc"] * 100, 3))
                            + " percent antibonding interaction below EFermi)"
                        )
                        if self.analysis_object.orbital_resolved:
                            text_orbital = self._generate_orbital_resolved_analysis_text(
                                orbital_resolved_data=properties,
                                type_pop=type_pop,
                                atom_name=str(type),
                                ion=item["ion"],
                            )
                            orb_info.extend(text_orbital)

                bonds = ",".join(bond_info[0:-1]) + ", and " + bond_info[-1] if len(bond_info) > 1 else bond_info[0]

                if len(orb_info) > 1:
                    orb_bonds = "".join(orb_info).replace(".In", ". In")
                else:
                    orb_bonds = orb_info[0] if orb_info else ""
                if item["env"] == "O:6":
                    self.text.append(
                        str(item["ion"])
                        + str(key + 1)
                        + " has an "
                        + str(self._coordination_environment_to_text(item["env"]))
                        + " coordination environment. It has "
                        + str(bonds)
                        + " bonds."
                    )
                    if orb_bonds:
                        self.text.append(orb_bonds)
                else:
                    self.text.append(
                        str(item["ion"])
                        + str(key + 1)
                        + " has a "
                        + str(self._coordination_environment_to_text(item["env"]))
                        + " coordination environment. It has "
                        + str(bonds)
                        + " bonds."
                    )
                    if orb_bonds:
                        self.text.append(orb_bonds)

        elif self.analysis_object.which_bonds == "all":
            relevant_ions = ", ".join(
                [
                    str(site.specie) + str(isite + 1)
                    for isite, site in enumerate(self.analysis_object.structure)
                    if isite in self.analysis_object.seq_ineq_ions
                ]
            )
            self.text = []
            self.text.append(
                "The compound "
                + str(self.condensed_bonding_analysis["formula"])
                + " has "
                + str(self.condensed_bonding_analysis["number_of_considered_ions"])
                + " symmetry-independent atoms(s) with relevant bonds: "
                + str(relevant_ions)
                + "."
            )

            for key, item in self.condensed_bonding_analysis["sites"].items():
                # It has 3 Ta-N (mean ICOHP: -4.78 eV, antibonding interactions below EFermi),
                bond_info = []
                orb_info = []
                for type, properties in item["bonds"].items():
                    if not properties["has_antibdg_states_below_Efermi"]:
                        bond_info.append(
                            str(properties["number_of_bonds"])
                            + " "
                            + item["ion"]
                            + "-"
                            + str(type)
                            + f" (mean I{type_pop}: "
                            ""
                            + properties[f"I{type_pop}_mean"]
                            + f"{units}, 0.0 percent antibonding interaction below EFermi)"
                        )
                        if self.analysis_object.orbital_resolved:
                            text_orbital = self._generate_orbital_resolved_analysis_text(
                                orbital_resolved_data=properties,
                                type_pop=type_pop,
                                atom_name=str(type),
                                ion=item["ion"],
                            )
                            orb_info.extend(text_orbital)
                    else:
                        bond_info.append(
                            str(properties["number_of_bonds"])
                            + " "
                            + item["ion"]
                            + "-"
                            + str(type)
                            + f" (mean I{type_pop}: "
                            ""
                            + properties[f"I{type_pop}_mean"]
                            + f"{units}, "
                            + str(round(properties["antibonding"]["perc"] * 100, 3))
                            + " percent antibonding interaction below EFermi)"
                        )

                        if self.analysis_object.orbital_resolved:
                            text_orbital = self._generate_orbital_resolved_analysis_text(
                                orbital_resolved_data=properties,
                                type_pop=type_pop,
                                atom_name=str(type),
                                ion=item["ion"],
                            )
                            orb_info.extend(text_orbital)

                bonds = ",".join(bond_info[0:-1]) + ", and " + bond_info[-1] if len(bond_info) > 1 else bond_info[0]

                if len(orb_info) > 1:
                    orb_bonds = "".join(orb_info).replace(".In", ". In")
                else:
                    orb_bonds = orb_info[0] if orb_info else ""
                if item["env"] == "O:6":
                    self.text.append(
                        str(item["ion"])
                        + str(key + 1)
                        + " has an "
                        + str(self._coordination_environment_to_text(item["env"]))
                        + " coordination environment. It has "
                        + str(bonds)
                        + " bonds."
                    )
                    if orb_bonds:
                        self.text.append(orb_bonds)
                else:
                    self.text.append(
                        str(item["ion"])
                        + str(key + 1)
                        + " has a "
                        + str(self._coordination_environment_to_text(item["env"]))
                        + " coordination environment. It has "
                        + str(bonds)
                        + " bonds."
                    )
                    if orb_bonds:
                        self.text.append(orb_bonds)

        if (
            "madelung_energy" in self.analysis_object.condensed_bonding_analysis
            and self.analysis_object.condensed_bonding_analysis["madelung_energy"] is not None
        ):
            self.text.append(
                "The Madelung energy of this crystal structure per unit cell is: "
                + str(self.analysis_object.condensed_bonding_analysis["madelung_energy"])
                + " eV."
            )

    def _generate_orbital_resolved_analysis_text(
        self,
        orbital_resolved_data: dict,
        ion: str,
        atom_name: str,
        type_pop: str,
    ):
        """
        Generate text from orbital-resolved analysis data of the most relevant COHP, COOP, or COBI.

        :param orbital_resolved_data: dict of orbital data from condensed bonding analysis object
        :param ion: name of ion at the site
        :param atom_name: name of atomic speice to which ion is bonded
        :param type_pop: population type analysed could be "COHP" or "COOP" or "COBI"

        Returns:
            A python list with text describing the orbital which contributes
            the most to the bonding and antibonding in the bond at site
        """
        orb_info = []
        if orbital_resolved_data["orbital_data"]["orbital_summary_stats"]:
            orb_names = []
            orb_contri = []
            # get atom-pair list with ion placed first
            atom_pair = self.analysis_object._sort_name([ion, atom_name], nameion=ion)
            if "max_bonding_contribution" in orbital_resolved_data["orbital_data"]["orbital_summary_stats"]:
                for orb, data in orbital_resolved_data["orbital_data"]["orbital_summary_stats"][
                    "max_bonding_contribution"
                ].items():
                    atom_pair_with_orb_name = self.analysis_object._sort_orbital_atom_pair(
                        atom_pair=atom_pair,
                        complete_cohp=self.analysis_object.chemenv.completecohp,
                        label=orbital_resolved_data["orbital_data"]["relevant_bonds"][0],
                        orb_pair=orb,
                    )
                    orb_names.append("-".join(atom_pair_with_orb_name))
                    orb_contri.append(
                        str(
                            round(
                                data * 100,
                                3,
                            )
                        )
                    )
            orb_names_anti = []
            orb_antibonding = []
            if "max_antibonding_contribution" in orbital_resolved_data["orbital_data"]["orbital_summary_stats"]:
                for orb, data in orbital_resolved_data["orbital_data"]["orbital_summary_stats"][
                    "max_antibonding_contribution"
                ].items():
                    atom_pair_with_orb_name = self.analysis_object._sort_orbital_atom_pair(
                        atom_pair=atom_pair,
                        complete_cohp=self.analysis_object.chemenv.completecohp,
                        label=orbital_resolved_data["orbital_data"]["relevant_bonds"][0],
                        orb_pair=orb,
                    )
                    orb_names_anti.append("-".join(atom_pair_with_orb_name))
                    orb_antibonding.append(
                        str(
                            round(
                                data * 100,
                                3,
                            )
                        )
                    )
            if len(orb_contri) > 1:
                orb_name_contri = ""
                for inx, name in enumerate(orb_names):
                    if len(orb_contri) == 2 and inx + 1 != len(orb_contri):
                        orb_name_contri += f"{name} "
                    elif 2 < len(orb_contri) != inx + 1:
                        orb_name_contri += f"{name}, "
                    else:
                        orb_name_contri += f"and {name}"

                orb_name_contri += " orbitals, contributing "
                for inx, contribution in enumerate(orb_contri):
                    if len(orb_contri) == 2 and inx + 1 != len(orb_contri):
                        orb_name_contri += f"{contribution} "
                    elif 2 < len(orb_contri) != inx + 1:
                        orb_name_contri += f"{contribution}, "
                    else:
                        orb_name_contri += f"and {contribution} percent, respectively"
                num_bonds = len(orbital_resolved_data["orbital_data"]["relevant_bonds"])
                bonds = "bonds" if num_bonds > 1 else "bond"
                orb_info.append(
                    f"In the {num_bonds} "
                    + "-".join(atom_pair)
                    + f" {bonds}, relative to the summed I{type_pop}s, "
                    + "the maximum bonding contribution is from "
                    + orb_name_contri
                )
            elif not orb_contri:
                num_bonds = len(orbital_resolved_data["orbital_data"]["relevant_bonds"])
                bonds = "bonds" if num_bonds > 1 else "bond"
                orb_info.append(
                    f"In the {num_bonds} "
                    + "-".join(atom_pair)
                    + f" {bonds}, relative to the summed I{type_pop}s, "
                    + f"no orbital has a bonding contribution greater than "
                    f"{self.analysis_object.orbital_cutoff*100} percent"
                )
            else:
                num_bonds = len(orbital_resolved_data["orbital_data"]["relevant_bonds"])
                bonds = "bonds" if num_bonds > 1 else "bond"
                orb_info.append(
                    f"In the {num_bonds} "
                    + "-".join(atom_pair)
                    + f" {bonds}, relative to the summed I{type_pop}s, "
                    + "the maximum bonding contribution is from the "
                    + f"{orb_names[0]}"
                    + f" orbital, contributing {orb_contri[0]} percent"
                )

            if len(orb_antibonding) > 1:
                orb_anti = ""
                for inx, name in enumerate(orb_names_anti):
                    if len(orb_names_anti) == 2 and inx + 1 != len(orb_names_anti):
                        orb_anti += f"{name} "
                    elif 2 < len(orb_antibonding) != inx + 1:
                        orb_anti += f"{name}, "
                    else:
                        orb_anti += f"and {name}"

                orb_anti += " orbitals, contributing "
                for inx, contribution in enumerate(orb_antibonding):
                    if len(orb_names_anti) == 2 and inx + 1 != len(orb_names_anti):
                        orb_anti += f"{contribution} "
                    elif 2 < len(orb_antibonding) != inx + 1:
                        orb_anti += f"{contribution}, "
                    else:
                        orb_anti += f"and {contribution} percent, respectively."
                orb_info.append(f", whereas the maximum antibonding contribution is from {orb_anti}")
            elif not orb_antibonding:
                orb_info.append(", whereas no significant antibonding contribution is found in this bond.")
            else:
                orb_info.append(
                    f", whereas the maximum antibonding contribution is from the "
                    f"{orb_names_anti[0]} orbital, contributing {orb_antibonding[0]} percent."
                )
        else:
            # get atom-pair list with ion placed first
            atom_pair = self.analysis_object._sort_name([ion, atom_name], nameion=ion)
            percentage_cutoff = round(self.analysis_object.orbital_cutoff * 100, 2)
            orb_info.append(
                f"No individual orbital interactions detected above {percentage_cutoff} percent"
                f" with summed I{type_pop} as reference for the " + "-".join(atom_pair) + " bond."
            )

        return orb_info

    def plot_cohps(
        self,
        xlim: list[float] | None = None,
        ylim: list[float] | None = [-4, 2],
        integrated: bool = False,
        title: str = "",
        save: bool = False,
        filename: str | None = None,
        sigma: float | None = None,
        hide: bool = False,
    ):
        """
        Automatically generate plots of the most relevant COHPs, COOPs, or COBIs.

        :param save: will save the plot to a file
        :param filename: name of the file to save the plot.
        :param ylim: energy scale that is shown in plot (eV)
        :param xlim: energy range for COHPs in eV
        :param integrated: if True, integrated COHPs will be shown
        :param sigma: Standard deviation of Gaussian broadening applied to
            population data. If None, no broadening will be added.
        :param title: sets the title of figure generated
        :param hide: if True, the plot will not be shown.

        Returns:
            A matplotlib object.

        """
        seq_cohps = self.analysis_object.seq_cohps
        if self.analysis_object.which_bonds == "cation-anion":
            seq_ineq_cations = self.analysis_object.seq_ineq_ions
        elif self.analysis_object.which_bonds == "all":
            seq_ineq_cations = self.analysis_object.seq_ineq_ions
        seq_labels = self.analysis_object.seq_labels_cohps
        structure = self.analysis_object.structure

        if len(seq_ineq_cations) >= 20:
            warnings.warn(
                "We will switch of displaying all plots "
                "as there are more than 20 inequivalent ions. "
                "We will instead save them in files called "
                "'automatic-analysis-*.png'."
            )
            hide = True
            save = True
            if filename is None:
                filename = "./automatic_analysis.png"

        for iplot, (ication, labels, cohps) in enumerate(zip(seq_ineq_cations, seq_labels, seq_cohps)):
            namecation = str(structure[ication].specie)

            cp = PlainCohpPlotter(
                are_coops=self.analysis_object.are_coops,
                are_cobis=self.analysis_object.are_cobis,
            )
            for label, cohp in zip(labels, cohps):
                if label is not None:
                    cp.add_cohp(namecation + str(ication + 1) + ": " + label, cohp)

            plot = cp.get_plot(integrated=integrated, sigma=sigma)
            plot.ylim(ylim)
            if xlim is not None:
                plot.xlim(xlim)

            plot.title(title)
            if save:
                if len(seq_ineq_cations) > 1:
                    if isinstance(filename, str):
                        filename = Path(filename)  # type: ignore
                    filename_new = (
                        filename.parent / f"{filename.stem}-{iplot}{filename.suffix}"  # type: ignore
                    )
                else:
                    filename_new = filename
                plot.savefig(filename_new)
                if hide:
                    plot.close()
        if not hide:
            plot.show()

    def plot_interactive_cohps(
        self,
        ylim: list[float] | None = None,
        xlim: list[float] | None = None,
        save_as_html: bool = False,
        filename: str | None = None,
        integrated: bool = False,
        title: str = "",
        sigma: float | None = None,
        label_resolved: bool = False,
        orbital_resolved: bool = False,
        hide: bool = False,
    ):
        """
        Automatically generate interactive plots of the most relevant COHPs, COBIs or COOPs.

        :param save_as_html: will save the plot to a html file
        :param filename: name of the file to save the plot.
        :param ylim: energy scale that is shown in plot (eV)
        :param xlim: energy range for COHPs in eV
        :param integrated: if True, integrated COHPs will be shown
        :param sigma: Standard deviation of Gaussian broadening applied to
            population data. If None, no broadening will be added.
        :param label_resolved: if true, relevant cohp curves will be further resolved based on band labels
        :param orbital_resolved: if true, relevant orbital interactions in cohp curves will be added to figure
        :param title: Title of the interactive plot
        :param hide: if True, the plot will not be shown.

        Returns:
            A plotly.graph_objects.Figure object.
        """
        cba_cohp_plot_data = {}  # Initialize dict to store plot data
        set_cohps = self.analysis_object.seq_cohps
        set_labels_cohps = self.analysis_object.seq_labels_cohps
        set_inequivalent_cations = self.analysis_object.seq_ineq_ions
        structure = self.analysis_object.structure

        for _iplot, (ication, labels, cohps) in enumerate(zip(set_inequivalent_cations, set_labels_cohps, set_cohps)):
            label_str = f"{structure[ication].specie!s}{ication + 1!s}: "
            for label, cohp in zip(labels, cohps):
                if label is not None:
                    cba_cohp_plot_data[label_str + label] = cohp

        ip = InteractiveCohpPlotter(
            are_coops=self.analysis_object.are_coops,
            are_cobis=self.analysis_object.are_cobis,
        )
        if label_resolved or orbital_resolved:
            ip.add_all_relevant_cohps(
                analyse=self.analysis_object,
                label_resolved=label_resolved,
                orbital_resolved=orbital_resolved,
            )
        else:
            ip.add_cohps_from_plot_data(plot_data_dict=cba_cohp_plot_data)

        plot = ip.get_plot(integrated=integrated, xlim=xlim, ylim=ylim, sigma=sigma)

        plot.update_layout(title_text=title)
        if save_as_html:
            plot.write_html(filename, include_mathjax="cdn")
        if not hide:
            return plot.show()

        return plot

    @staticmethod
    def _coordination_environment_to_text(ce: str):
        """
        Convert a coordination environment string into a text description of the environment.

        :param ce: output from ChemEnv package (e.g., "O:6")

        Returns:
            A text description of coordination environment
        """
        if ce == "S:1":
            return "single (CN=1)"
        if ce == "L:2":
            return "linear (CN=2)"
        if ce == "A:2":
            return "angular (CN=2)"
        if ce == "TL:3":
            return "trigonal planar (CN=3)"
        if ce == "TY:3":
            return "triangular non-coplanar (CN=3)"
        if ce == "TS:3":
            return "t-shaped (CN=3)"
        if ce == "T:4":
            return "tetrahedral (CN=4)"
        if ce == "S:4":
            return "square planar (CN=4)"
        if ce == "SY:4":
            return "square non-coplanar (CN=4)"
        if ce == "SS:4":
            return "see-saw like (CN=4)"
        if ce == "PP:5":
            return "pentagonal (CN=5)"
        if ce == "S:5":
            return "square pyramidal (CN=5)"
        if ce == "T:5":
            return "trigonal bipyramidal (CN=5)"
        if ce == "O:6":
            return "octahedral (CN=6)"
        if ce == "T:6":
            return "trigonal prismatic (CN=6)"
        if ce == "PP:6":
            return "pentagonal pyramidal (CN=6)"
        if ce == "PB:7":
            return "pentagonal bipyramidal (CN=7)"
        if ce == "ST:7":
            return "square-face capped trigonal prismatic (CN=7)"
        if ce == "ET:7":
            return "end-trigonal-face capped trigonal prismatic (CN=7)"
        if ce == "FO:7":
            return "face-capped octahedron (CN=7)"
        if ce == "C:8":
            return "cubic (CN=8)"
        if ce == "SA:8":
            return "square antiprismatic (CN=8)"
        if ce == "SBT:8":
            return "square-face bicapped trigonal prismatic (CN=8)"
        if ce == "TBT:8":
            return "triangular-face bicapped trigonal prismatic (CN=8)"
        if ce == "DD:8":
            return "dodecahedronal (with triangular faces) (CN=8)"
        if ce == "DDPN:8":
            return "dodecahedronal (with triangular faces - p2345 plane normalized) (CN=8)"
        if ce == "HB:8":
            return "hexagonal bipyramidal (CN=8)"
        if ce == "BO_1:8":
            return "bicapped octahedral (opposed cap faces) (CN=8)"
        if ce == "BO_2:8":
            return "bicapped octahedral (cap faces with one atom in common) (CN=8)"
        if ce == "BO_3:8":
            return "bicapped octahedral (cap faces with one edge in common) (CN=8)"
        if ce == "TC:9":
            return "triangular cupola (CN=9)"
        if ce == "TT_1:9":
            return "Tricapped triangular prismatic (three square - face caps) (CN=9)"
        if ce == "TT_2:9":
            return "Tricapped triangular prismatic (two square - face caps and one triangular - face cap) (CN=9)"
        if ce == "TT_3:9":
            return "Tricapped triangular prism (one square - face cap and two triangular - face caps) (CN=9)"
        if ce == "HD:9":
            return "Heptagonal dipyramidal (CN=9)"
        if ce == "TI:9":
            return "tridiminished icosohedral (CN=9)"
        if ce == "SMA:9":
            return "Square-face monocapped antiprism (CN=9)"
        if ce == "SS:9":
            return "Square-face capped square prismatic (CN=9)"
        if ce == "TO_1:9":
            return "Tricapped octahedral (all 3 cap faces share one atom) (CN=9)"
        if ce == "TO_2:9":
            return "Tricapped octahedral (cap faces are aligned) (CN=9)"
        if ce == "TO_3:9":
            return "Tricapped octahedron (all 3 cap faces are sharing one edge of a face) (CN=9)"
        if ce == "PP:10":
            return "Pentagonal prismatic (CN=10)"
        if ce == "PA:10":
            return "Pentagonal antiprismatic (CN=10)"
        if ce == "SBSA:10":
            return "Square-face bicapped square antiprismatic (CN=10)"
        if ce == "MI:10":
            return "Metabidiminished icosahedral (CN=10)"
        if ce == "S:10":
            return "sphenocoronal (CN=10)"
        if ce == "H:10":
            return "Hexadecahedral (CN=10)"
        if ce == "BS_1:10":
            return "Bicapped square prismatic (opposite faces) (CN=10)"
        if ce == "BS_1:10":
            return "Bicapped square prismatic (opposite faces) (CN=10)"
        if ce == "BS_2:10":
            return "Bicapped square prism(adjacent faces) (CN=10)"
        if ce == "TBSA:10":
            return "Trigonal-face bicapped square antiprismatic (CN=10)"
        if ce == "PCPA:11":
            return "Pentagonal - face capped pentagonal antiprismatic (CN=11)"
        if ce == "H:11":
            return "Hendecahedral (CN=11)"
        if ce == "SH:11":
            return "Sphenoid hendecahedral (CN=11)"
        if ce == "CO:11":
            return "Cs - octahedral (CN=11)"
        if ce == "DI:11":
            return "Diminished icosahedral (CN=12)"
        if ce == "I:12":
            return "Icosahedral (CN=12)"
        if ce == "PBP: 12":
            return "Pentagonal - face bicapped pentagonal prismatic (CN=12)"
        if ce == "TT:12":
            return "Truncated tetrahedral (CN=12)"
        if ce == "C:12":
            return "Cuboctahedral (CN=12)"
        if ce == "AC:12":
            return "Anticuboctahedral (CN=12)"
        if ce == "SC:12":
            return "Square cupola (CN=12)"
        if ce == "S:12":
            return "Sphenomegacorona (CN=12)"
        if ce == "HP:12":
            return "Hexagonal prismatic (CN=12)"
        if ce == "HA:12":
            return "Hexagonal antiprismatic (CN=12)"
        if ce == "SH:13":
            return "Square-face capped hexagonal prismatic (CN=13)"
        if ce == "1":
            return "1-fold"
        if ce == "2":
            return "2-fold"
        if ce == "3":
            return "3-fold"
        if ce == "4":
            return "4-fold"
        if ce == "5":
            return "5-fold"
        if ce == "6":
            return "6-fold"
        if ce == "7":
            return "7-fold"
        if ce == "8":
            return "8-fold"
        if ce == "9":
            return "9-fold"
        if ce == "10":
            return "10-fold"
        if ce == "11":
            return "11-fold"
        if ce == "12":
            return "12-fold"
        if ce == "13":
            return "13-fold"
        if ce == "14":
            return "14-fold"
        if ce == "15":
            return "15-fold"
        if ce == "16":
            return "16-fold"
        if ce == "17":
            return "17-fold"
        if ce == "18":
            return "18-fold"
        if ce == "19":
            return "19-fold"
        if ce == "20":
            return "20-fold"
        if ce == "21":
            return "21-fold"
        if ce == "22":
            return "22-fold"
        if ce == "23":
            return "23-fold"
        if ce == "24":
            return "24-fold"
        if ce == "25":
            return "25-fold"
        if ce == "26":
            return "26-fold"
        if ce == "27":
            return "27-fold"
        if ce == "28":
            return "28-fold"
        if ce == "29":
            return "29-fold"
        if ce == "30":
            return "30-fold"
        return ce

    def write_description(self):
        """Print the description of the COHPs or COBIs or COOPs to the screen."""
        for textpart in self.text:
            print(textpart)

    @staticmethod
    def get_calc_quality_description(quality_dict):
        """
        Generate a text description of the LOBSTER calculation quality.

        :param quality_dict: python dictionary from lobsterpy.analysis.get_lobster_calc_quality_summary
        """
        text_des = []

        for key, val in quality_dict.items():
            if key == "minimal_basis":
                if val:
                    text_des.append("The LOBSTER calculation used minimal basis.")
                if not val:
                    text_des.append(
                        "Consider rerunning the calculation with the minimum basis as well. Choosing a "
                        "larger basis set is only recommended if you see a significant improvement of "
                        "the charge spilling."
                    )

            elif key == "charge_spilling":
                text_des.append(
                    "The absolute and total charge spilling for the calculation is {} and {} %, "
                    "respectively.".format(
                        quality_dict[key]["abs_charge_spilling"],
                        quality_dict[key]["abs_total_spilling"],
                    )
                )
            elif key == "band_overlaps_analysis":
                if quality_dict[key]["file_exists"]:
                    if quality_dict[key]["has_good_quality_maxDeviation"]:
                        text_des.append(
                            "The bandOverlaps.lobster file is generated during the LOBSTER run. This "
                            "indicates that the projected wave function is not completely orthonormalized; "
                            "however, the maximal deviation values observed compared to the identity matrix "
                            "is below the threshold of 0.1."
                        )
                    else:
                        text_des.append(
                            "The bandOverlaps.lobster file is generated during the LOBSTER run. This "
                            "indicates that the projected wave function is not completely orthonormalized. "
                            "The maximal deviation value from the identity matrix is {}, and there are "
                            "{} percent k-points above the deviation threshold of 0.1. Please check the "
                            "results of other quality checks like dos comparisons, charges, "
                            "charge spillings before using the results for further "
                            "analysis.".format(
                                quality_dict[key]["max_deviation"],
                                quality_dict[key]["percent_kpoints_abv_limit"],
                            )
                        )
                else:
                    text_des.append(
                        "The projected wave function is completely orthonormalized as no "
                        "bandOverlaps.lobster file is generated during the LOBSTER run."
                    )

            elif key == "charge_comparisons":
                if val:
                    for charge in ["mulliken", "loewdin"]:
                        if val[f"bva_{charge}_agree"]:
                            text_des.append(
                                f"The atomic charge signs from {charge.capitalize()} population analysis "
                                f"agree with the bond valence analysis."
                            )
                        if not val[f"bva_{charge}_agree"]:
                            text_des.append(
                                f"The atomic charge signs from {charge.capitalize()} population analysis "
                                f"do not agree with the bond valence analysis."
                            )
                else:
                    text_des.append(
                        "Oxidation states from BVA analyzer cannot be determined. "
                        "Thus BVA charge comparison is not conducted."
                    )

            elif key == "dos_comparisons":
                comp_types = []
                tani_index = []
                for orb in val:
                    if orb.split("_")[-1] in ["s", "p", "d", "f", "summed"]:
                        comp_types.append(orb.split("_")[-1])
                        tani_index.append(str(val[orb]))
                text_des.append(
                    "The Tanimoto index from DOS comparisons in the energy range between {}, {} eV "
                    "for {} orbitals are: {}.".format(
                        val["e_range"][0],
                        val["e_range"][1],
                        ", ".join(comp_types),
                        ", ".join(tani_index),
                    )
                )

        return text_des

    @staticmethod
    def write_calc_quality_description(calc_quality_text):
        """Print the calculation quality description to the screen."""
        print(" ".join(calc_quality_text))



"""
Dictionaries for plotly figure layouts.

Layout inspired by
https://github.com/materialsproject/crystaltoolkit/blob/main/crystal_toolkit/components/bandstructure.py

Contains dicts for
    1. general figure layout
    2. axes ((I)COHP / energy)
    3. traces (Spin.up / Spin.down)
    4. legend

"""

layout_dict = {
    "titlefont": {"size": 18},
    "showlegend": True,
    "title_x": 0.1,
    "title_y": 0.9,
    "hovermode": "closest",
    "paper_bgcolor": "rgba(0,0,0,0)",
    "plot_bgcolor": "rgba(0,0,0,0)",
}

"""
    General layout of Plotly figure.

    :param dict titlefont: Line style dictionary.
    :param bool showlegend: Legend hide or show.
    :param float title_x: x axis title size.
    :param float title_y: y axis title size.
    :param str hovermode: hover behaviour.
    :param str paper_bgcolor: background color.
    :param str plot_bgcolor: plot background color.
"""

cohp_axis_style_dict = {
    "titlefont": {"size": 20},
    "tickfont": {"size": 16},
    "ticks": "inside",
    "tickwidth": 2,
    "showgrid": False,
    "showline": True,
    "zeroline": True,
    "zerolinecolor": "black",
    "zerolinewidth": 3,
    "mirror": True,
    "linewidth": 2,
    "linecolor": "black",
}

"""
    COXX axis style.

    :param dict titlefont: axis title font style.
    :param dict tickfont: axis tick font style.
    :param str ticks: ticks style.
    :param float tickwidth: width of ticks.
    :param bool showgrid: show or hide axis grid.
    :param bool showline: show or hide axis.
    :param bool zeroline: show or hide zero line.
    :param str zerolinecolor: color of zero line.
    :param float zerolinewidth: width of zero line.
    :param bool mirror: mirror axis.
    :param float linewidth: width of axis line.
    :param str linecolor: color of axis line.
"""

energy_axis_style_dict = {
    "titlefont": {"size": 20},
    "tickfont": {"size": 16},
    "ticks": "inside",
    "tickwidth": 2,
    "showgrid": False,
    "showline": True,
    "zeroline": True,
    "zerolinecolor": "rgba(0,0,0,0.5)",
    "zerolinewidth": 3,
    "mirror": True,
    "linewidth": 2,
    "linecolor": "black",
}

"""
    Energy axis style.

    :param dict titlefont: axis title font style.
    :param dict tickfont: axis tick font style.
    :param str ticks: ticks style.
    :param float tickwidth: width of ticks.
    :param bool showgrid: show or hide axis grid.
    :param bool showline: show or hide axis.
    :param bool zeroline: show or hide zero line.
    :param str zerolinecolor: color of zero line.
    :param float zerolinewidth: width of zero line.
    :param bool mirror: mirror axis.
    :param float linewidth: width of axis line.
    :param str linecolor: color of axis line.
"""

spin_up_trace_style_dict = {
    "line": {"width": 3},
    "mode": "lines",
    "visible": "legendonly",
}

"""
    Line style for Spin.up traces.

    :param dict line: Line style dictionary.
    :param str mode: Plotly mode.
    :param str visible: Trace visibility.
"""

spin_down_trace_style_dict = {
    "line": {"width": 3, "dash": "dash"},
    "mode": "lines",
    "visible": "legendonly",
}

"""
    Line style for Spin.down traces.

    :param dict line: Line style dictionary.
    :param str mode: Plotly mode.
    :param str visible: Trace visibility.
"""

legend_style_dict = {
    "bordercolor": "black",
    "borderwidth": 2,
    "traceorder": "normal",
}

"""
    Legend style.

    :param str bordercolor: border color of legend.
    :param int borderwidth: width of border.
    :param str traceorder: order of trace.
"""



# Copyright (c) lobsterpy development team
# Distributed under the terms of a BSD 3-Clause "New" or "Revised" License

"""Script to analyze Lobster outputs from the command line."""

from __future__ import annotations

import argparse
import json
import os
from math import log, sqrt
from pathlib import Path

import matplotlib.style
from monty.os.path import zpath
from pymatgen.electronic_structure.cohp import CompleteCohp
from pymatgen.io.lobster import Icohplist

from lobsterpy.cohp.analyze import Analysis
from lobsterpy.cohp.describe import Description
from lobsterpy.featurize.utils import get_file_paths
from lobsterpy.plotting import (
    IcohpDistancePlotter,
    PlainCohpPlotter,
    PlainDosPlotter,
    get_style_list,
)


def main() -> None:
    """Entry point for setup.py installer."""
    args = get_parser().parse_args()
    run(args)


def get_parser() -> argparse.ArgumentParser:
    """Construct argumentparser with subcommands and sections."""
    parser = argparse.ArgumentParser(description="Analyze and plot results from Lobster runs.")

    # Arguments that are needed by different actions, but not always

    incar_file = argparse.ArgumentParser(add_help=False)
    incar_file.add_argument(
        "-fincar",
        "--file-incar",
        default="INCAR",
        dest="incar",
        type=Path,
        help='path to INCAR. Default is "INCAR".',
    )

    charge_file = argparse.ArgumentParser(add_help=False)
    charge_file.add_argument(
        "-fcharge",
        "--file-charge",
        default="CHARGE.lobster",
        dest="charge",
        type=Path,
        help='path to CHARGE.lobster. Default is "CHARGE.lobster"',
    )

    structure_file = argparse.ArgumentParser(add_help=False)
    structure_file.add_argument(
        "-fstruct",
        "--file-structure",
        default="POSCAR",
        dest="structure",
        type=Path,
        help='path to structure file. Default is "POSCAR". '
        'Can also read "POSCAR.lobster" file or any '
        'suitable file format supported by pymatgen "Structure.from_file" method.',
    )

    potcar_file = argparse.ArgumentParser(add_help=False)
    potcar_file.add_argument(
        "-fpotcar",
        "--file-potcar",
        default="POTCAR",
        dest="potcar",
        type=Path,
        help='path to POTCAR. Default is "POTCAR".',
    )

    icohplist_file = argparse.ArgumentParser(add_help=False)
    icohplist_file.add_argument(
        "-ficohp",
        "--file-icohplist",
        default="ICOHPLIST.lobster",
        dest="icohplist",
        type=Path,
        help='path to ICOHPLIST.lobster. Default is "ICOHPLIST.lobster"',
    )

    coxxcar_file = argparse.ArgumentParser(add_help=False)
    coxxcar_file.add_argument(
        "-fcohp",
        "--file-cohpcar",
        default="COHPCAR.lobster",
        dest="cohpcar",
        type=Path,
        help='path to COHPCAR.lobster. Default is "COHPCAR.lobster". This argument '
        "can also read COBICARs or COOPCARs. One needs to use appropriate --cobis or "
        "--coops options along with this argument when plotting",
    )

    doscar_file = argparse.ArgumentParser(add_help=False)
    doscar_file.add_argument(
        "-fdos",
        "--file-doscar",
        default="DOSCAR.lobster",
        dest="doscar",
        type=Path,
        help='path to DOSCAR.lobster. Default is "DOSCAR.lobster".',
    )

    user_basis_arg = argparse.ArgumentParser(add_help=False)
    user_basis_arg.add_argument(
        "--userbasis",
        "--user-basis",
        default=None,
        type=_element_basis,
        nargs="+",
        help="Use the specific basis provided by the user to "
        "generate the inputs (e.g.,  --userbasis Cr.3d.3p.4s N.2s.2p). "
        "Default is None.",
    )

    # groups of arguments for specific actions

    calc_quality_description_file_parent = argparse.ArgumentParser(add_help=False)

    # Input args for specifically needed for calculation quality description
    calc_quality_description_file_group = calc_quality_description_file_parent.add_argument_group("")
    calc_quality_description_file_group.add_argument(
        "-fvasprun",
        "--file-vasprun",
        default="vasprun.xml",
        dest="vasprun",
        type=Path,
        help='path to vasprun.xml. Default is "vasprun.xml".',
    )
    calc_quality_description_file_group.add_argument(
        "-fbandoverlaps",
        "--file-bandoverlaps",
        default="bandOverlaps.lobster",
        dest="bandoverlaps",
        type=Path,
        help='path to bandOverlaps.lobster. Default is "bandOverlaps.lobster".',
    )
    calc_quality_description_file_group.add_argument(
        "-flobsterin",
        "--file-lobsterin",
        default="lobsterin",
        dest="lobsterin",
        type=Path,
        help='path to lobsterin. Default is "lobsterin".',
    )
    calc_quality_description_file_group.add_argument(
        "-flobsterout",
        "--file-lobsterout",
        default="lobsterout",
        dest="lobsterout",
        type=Path,
        help='path to lobsterout. Default is "lobsterout".',
    )
    calc_quality_description_file_group.add_argument(
        "-potsymbols",
        "--potcar-symbols",
        dest="potcarsymbols",
        type=_potcar_symbols,
        # nargs="+",
        help="List of potcar symbols",
    )

    # group of arguments related to writing LOBSTER calcs inputs
    output_parent = argparse.ArgumentParser(add_help=False)
    output_file_group = output_parent.add_argument_group("Output files")
    output_file_group.add_argument(
        "-fincarout",
        "--file-incar-out",
        default="INCAR.lobsterpy",
        dest="incarout",
        type=Path,
        help='path to INCAR that lobsterpy generates. Default is "INCAR.lobsterpy"',
    )
    output_file_group.add_argument(
        "-flobsterin",
        "--file-lobsterin",
        default="lobsterin.lobsterpy",
        dest="lobsterinout",
        type=Path,
        help='Base for path to lobsterins that lobsterpy generates. Default is "lobsterin.lobsterpy"',
    )
    output_file_group.add_argument(
        "--overwrite",
        "--overwrite-files",
        dest="overwrite",
        default=False,
        action="store_true",
        help="overwrites already created INCARs and lobsterins with the give name.",
    )
    # TODO: Add some output arguments: options to supply your own basis
    # General matplotlib plotting arguments common to all kinds of plots

    plotting_parent = argparse.ArgumentParser(add_help=False)
    plotting_group = plotting_parent.add_argument_group("Plotting")
    plotting_group.add_argument(
        "--fontsize",
        "--font-size",
        type=float,
        dest="fontsize",
        default=None,
        help="Base font size",
    )
    plotting_group.add_argument(
        "-ht",
        "--height",
        type=float,
        dest="height",
        default=None,
        help="Plot height in inches",
    )
    plotting_group.add_argument(
        "--hideplot",
        "--hide-plot",
        dest="hideplot",
        action="store_true",
        help="Hide plot output. Especially relevant when plots are saved.",
    )
    plotting_group.add_argument(
        "-nbs",
        "--no-base-style",
        "--nobasestyle",
        action="store_true",
        dest="no_base_style",
        help=(
            "Disable inbuilt style entirely. This may prevent interference with external "
            "stylesheets when using --style."
        ),
    )
    plotting_group.add_argument(
        "-sty",
        "--style",
        type=str,
        nargs="+",
        dest="style",
        default=None,
        help="Matplotlib style sheet(s) for plot appearance",
    )
    plotting_group.add_argument(
        "--save-plot",
        "--saveplot",
        "-s",
        type=Path,
        metavar="FILENAME",
        default=None,
        dest="save_plot",
        help="Save plot to file",
    )
    plotting_group.add_argument("--title", type=str, default="", help="Plot title")
    plotting_group.add_argument(
        "-wd",
        "--width",
        type=float,
        dest="width",
        default=None,
        help="Plot width in inches",
    )
    plotting_group.add_argument(
        "-x",
        "--xlim",
        dest="xlim",
        nargs=2,
        default=None,
        type=float,
        help="Set x-axis limits for the plots",
    )
    plotting_group.add_argument(
        "-y",
        "--ylim",
        dest="ylim",
        nargs=2,
        default=None,
        type=float,
        help="Set y-axis limits for the plots",
    )
    broadening_group = plotting_group.add_mutually_exclusive_group()
    broadening_group.add_argument_group("Broadening")
    broadening_group.add_argument(
        "--fwhm",
        type=float,
        default=None,
        help="Full-width-half-maximum of Gaussian broadening.",
    )
    broadening_group.add_argument(
        "--sigma",
        type=float,
        default=None,
        help="Standard deviation of Gaussian broadening.",
    )
    # Group pf arguments specific to dos plotter
    dos_plotting_parent = argparse.ArgumentParser(add_help=False)
    dos_plotting_group = dos_plotting_parent.add_argument_group("Plotting")
    dos_plotting_group.add_argument(
        "-addtdos",
        "--addtotal",
        "--add-total",
        action="store_true",
        help="Add total dos to DOS plot.",
    )
    dos_plotting_group.add_argument(
        "-el",
        "--element",
        type=str,
        dest="element",
        nargs="+",
        default=None,
        help="Add spd DOS projections for requested element to the DOS plot",
    )
    dos_plotting_group.add_argument(
        "-orb",
        "--orbital",
        type=str,
        dest="orbital",
        nargs="+",
        default=None,
        help="Orbital name for the site for which DOS are to be added",
    )
    dos_plotting_group.add_argument(
        "--site",
        type=int,
        dest="site",
        nargs="+",
        default=None,
        help="Site index in the crystal structure for which DOS need to be added",
    )
    dos_plotting_group.add_argument(
        "-sspin",
        "--summedspins",
        "--summed-spins",
        dest="summedspins",
        action="store_true",
        help="Plot summed spins DOS",
    )

    group = dos_plotting_parent.add_mutually_exclusive_group()
    group.add_argument(
        "-eldos",
        "--elementdos",
        dest="elementdos",
        action="store_true",
        help="Add DOS projections for each element to the DOS plot",
    )
    group.add_argument(
        "--spddos",
        "--spd-dos",
        dest="spddos",
        action="store_true",
        help="Add spd projected dos to the DOS plot",
    )
    # Argument common to COHPs / COOPs / COBIs or DOS plots
    advanced_plotting_args = argparse.ArgumentParser(add_help=False)
    advanced_plotting_args.add_argument(
        "--invertaxis",
        "--invert-axis",
        dest="invertaxis",
        action="store_true",
        help="Invert plot axis of DOS or COOPs COHPs or COBIS",
    )
    # Argument specific to COHPs / COOPs / COBIs plots
    coxx_plotting_args = argparse.ArgumentParser(add_help=False)
    coxx_plotting_args.add_argument(
        "-integr",
        "--integrated",
        dest="integrated",
        action="store_true",
        help="Show integrated cohp/cobi/coop plots.",
    )
    # Arguments specific to lobsterpy.cohp.analyze.Analysis and
    # lobsterpy.cohp.describe.Description class
    auto_parent = argparse.ArgumentParser(add_help=False)
    auto_group = auto_parent.add_argument_group("Adjustable automatic analysis parameters")
    auto_group.add_argument(
        "-allb",
        "--allbonds",
        "--all-bonds",
        action="store_true",
        default=False,
        help="Consider all bonds during the automatic analysis, not only cation-anion bonds (default) ",
    )
    auto_group.add_argument(
        "--cutofficohp",
        type=float,
        default=0.1,
        help="Consider only bonds that are stronger than cutoff_icoxx * strongest ICOXX "
        " (ICOHP or ICOBI or ICOOP) for automatic analysis.",
    )
    auto_group.add_argument(
        "-fjson",
        "--file-json",
        nargs="?",
        type=Path,
        default=None,
        metavar="FILENAME",
        const=Path("lobsterpy.json"),
        help="Write a JSON file with the most important information",
    )
    auto_group.add_argument(
        "--noisecutoff",
        type=float,
        default=None,
        help="Sets the lower limit of icohps or icoops or icobis considered in automatic analysis",
    )
    auto_group.add_argument(
        "-orbresol",
        "--orbitalresolved",
        "--orbital-resolved",
        action="store_true",
        default=False,
        help="Switch on orbital resolved analysis of (I)COHPs or (I)COBIs or (I)COOPs with all relevant orbitals.",
    )
    auto_group.add_argument(
        "-orbcutoff",
        "--orbitalcutoff",
        "--orbital-cutoff",
        type=float,
        default=0.05,
        help="Consider only orbital interactions that are stronger than orbitalintcutoff * 100 of "
        "relevant bonds (ICOHP or ICOBI or ICOOP). Can only be used in combination "
        "with orbital-resolved automatic analysis (--orbitalresolved).",
    )

    auto_group.add_argument(
        "-sspins",
        "--summedspins",
        "--summed-spins",
        action="store_true",
        default=False,
        help="Sum COHP `Spin.up` and `Spin.down` populations for automatic analysis",
    )

    # Argument that will help to switch automatic analysis
    analysis_switch = argparse.ArgumentParser(add_help=False)
    analysis_group = analysis_switch.add_argument_group(
        "Arguments to switch type of files analyzed during automatic analysis"
        " (Also indicates file type for 'plot/plot-icohp-distance' action in cli)"
    )
    analysis_group.add_argument(
        "--cobis",
        "--cobis",
        action="store_true",
        help="Setting this option starts automatic bonding analysis using COBIs"
        " (Also indicates plotter that input file contains COBI data)",
    )
    analysis_group.add_argument(
        "--coops",
        "--coops",
        action="store_true",
        help="Setting this option starts automatic bonding analysis using COOPs"
        " (Also indicates plotter that input file contains COOP data)",
    )

    # Specific to interactive plotter args
    interactive_plotter_args = argparse.ArgumentParser(add_help=False)
    interactive_plotter_group = interactive_plotter_args.add_argument_group("Options specific to interactive plotter")
    interactive_plotter_group.add_argument(
        "-labresol",
        "--labelresolved",
        "--label-resolved",
        action="store_true",
        help="Create automatic interactive plots with all relevant bond labels. "
        "If not set, plots consist of summed cohps.",
    )

    # Args specific to calc quality description dict and texts
    calc_quality_args = argparse.ArgumentParser(add_help=False)
    calc_quality_args_group = calc_quality_args.add_argument_group(
        "Options to change default output of quality description"
    )
    calc_quality_args_group.add_argument(
        "--bvacomp",
        "--bva-comp",
        action="store_true",
        default=False,
        help="Enable the BVA charge comparison in automatic LOBSTER calc quality analysis ",
    )
    calc_quality_args_group.add_argument(
        "--doscomp",
        "--dos-comp",
        action="store_true",
        default=False,
        help="Enable the DOS comparison in automatic LOBSTER calc quality analysis ",
    )
    calc_quality_args_group.add_argument(
        "--erange",
        dest="erange",
        nargs=2,
        default=[-5, 0],
        type=int,
        help="Energy range for DOS comparisons",
    )
    calc_quality_args_group.add_argument(
        "-fcalcqualjson",
        "--file-calc-quality-json",
        nargs="?",
        type=Path,
        default=None,
        metavar="FILENAME",
        const=Path("calc_quality_json.json"),
        help="Write a JSON file with the LOBSTER calc quality analysis",
    )
    calc_quality_args_group.add_argument(
        "--nbins",
        dest="nbins",
        default=None,
        type=int,
        help="Number of bins for DOS comparisons",
    )
    # Build the actions using arguments defined earlier
    subparsers = parser.add_subparsers(
        dest="action",
        required=True,
        help="Use -h/--help after the chosen subcommand to see further options.",
    )
    subparsers.add_parser(
        "create-inputs",
        aliases=["createinputs"],
        parents=[
            incar_file,
            structure_file,
            potcar_file,
            user_basis_arg,
            output_parent,
        ],
        help="Create inputs for lobster computation. It works only with PBE POTCARs, "
        "as, currently, only the pbeVASPfit2015 basis fitted to PBE POTCARs in LOBSTER includes "
        "additional orbitals relevant for the solid-state materials. Please check out our "
        "publication https://doi.org/10.1002/cplu.202200123 and LOBSTER program manual for more information.",
    )
    subparsers.add_parser(
        "description",
        parents=[
            charge_file,
            coxxcar_file,
            icohplist_file,
            structure_file,
            auto_parent,
            analysis_switch,
        ],
        help=("Deliver a text description from automatic analysis of COHPs or COBIS or COOP results from Lobster run"),
    )
    subparsers.add_parser(
        "description-quality",
        parents=[
            calc_quality_description_file_parent,
            charge_file,
            doscar_file,
            potcar_file,
            structure_file,
            calc_quality_args,
        ],
        help=(
            "Deliver a text description of the LOBSTER calc quality analysis. "
            "Mandatory required files: POSCAR, POTCAR, lobsterout, lobsterin. "
            "Optional files (BVA comparison): CHARGE.lobster, "
            "(DOS comparison): DOSCAR.lobster/ DOSCAR.LSO.lobster, Vasprun.xml."
        ),
    )
    subparsers.add_parser(
        "plot-automatic",
        aliases=[
            "plot-auto",
            "automatic-plot",
            "automaticplot",
            "auto-plot",
            "autoplot",
        ],
        parents=[
            charge_file,
            coxxcar_file,
            icohplist_file,
            structure_file,
            auto_parent,
            plotting_parent,
            coxx_plotting_args,
            analysis_switch,
        ],
        help=(
            "Plot most important COHPs or COBIs or COOPs automatically."
            " This option also includes an automatic description."
        ),
    )
    subparsers.add_parser(
        "plot-automatic-ia",
        aliases=[
            "plot-auto-ia",
            "automatic-plot-ia",
            "automaticplotia",
            "autoplotia",
            "auto-plot-ia",
        ],
        parents=[
            charge_file,
            coxxcar_file,
            icohplist_file,
            structure_file,
            auto_parent,
            plotting_parent,
            coxx_plotting_args,
            interactive_plotter_args,
            analysis_switch,
        ],
        help=("Creates an interactive plot of most important COHPs or COBIs or COOPs automatically."),
    )
    subparsers.add_parser(
        "plot-dos",
        aliases=["plotdos"],
        parents=[
            doscar_file,
            structure_file,
            dos_plotting_parent,
            advanced_plotting_args,
            plotting_parent,
        ],
        help="Plots DOS from lobster computation.",
    )
    subparsers.add_parser(
        "plot-icohp-distance",
        aliases=["ploticohpdistance"],
        parents=[icohplist_file, plotting_parent, analysis_switch],
        help="Plot ICOHPs or ICOOPs or ICOBIs with respect to bond lengths",
    )
    # Mode for normal plotting (without automatic detection of relevant COHPs)
    plot_parser = subparsers.add_parser(
        "plot",
        parents=[
            coxxcar_file,
            structure_file,
            plotting_parent,
            advanced_plotting_args,
            coxx_plotting_args,
            analysis_switch,
        ],
        help="Plot specific COHPs/COBIs/COOPs based on bond numbers.",
    )
    plot_parser.add_argument(
        "bond_numbers",
        nargs="+",
        type=int,
        help="List of bond numbers, determining COHPs/COBIs/COOPs to include in plot.",
    )
    plot_grouping = plot_parser.add_mutually_exclusive_group()
    plot_grouping.add_argument("--multi_cobis", action="store_true", help="Plot COBICARs including multi-center bonds.")

    plot_grouping.add_argument(
        "--summed",
        action="store_true",
        help="Show a summed COHP",
    )
    plot_grouping.add_argument(
        "-orbwise",
        "--orbitalwise",
        dest="orbitalwise",
        nargs="+",
        default=None,
        type=str,
        help=(
            "Plot cohps of specific orbitals. e.g. to plot 2s-2s interaction of "
            'bond with label 1, use "lobsterpy plot 1 --orbitalwise 2s-2s". '
            'To plot all orbitalwise cohps of one bond, you can use "all" instead of "2s-2s". '
            "To plot orbitalwise interactions of more than one bond, use, for example, "
            '"lobsterpy plot 1 1 --orbitalwise "3s-3s" "2px-3s"'
        ),
    )
    return parser


def _element_basis(string: str):
    """
    Parse element and basis from string.

    :param string: string to parse

    Returns:
            element, basis
    """
    cut_list = string.split(".")
    element = cut_list[0]
    basis = " ".join(cut_list[1:])
    return element, basis


def _potcar_symbols(string: str):
    """
    Parse string of potcar symbols and return a list.

    :param string: string of potcar symbols

    Returns:
        list of potcar symbols
    """
    return string.split(" ")


def _user_figsize(width, height, aspect=None):
    """Get figsize options from user input, if any.

    If only width x or height is provided, use a target aspect ratio to derive
    the other one.

    Returns a dict which can be merged into style kwargs
    """
    if width is None and height is None:
        return {}
    if width is not None and height is not None:
        return {"figure.figsize": (width, height)}

    if aspect is None:
        aspect = (sqrt(5) + 1) / 2  # Golden ratio
    if width is None:
        return {"figure.figsize": (height * aspect, height)}
    return {"figure.figsize": (width, width / aspect)}


# TODO: add automatic functionality for COBIs, COOPs
def run(args):
    """
    Run actions based on args.

    :param args: args for cli

    """
    if args.action in [
        "plot-auto",
        "automatic-plot",
        "automaticplot",
        "auto-plot",
        "autoplot",
    ]:
        args.action = "plot-automatic"

    if args.action in [
        "plot-auto-ia",
        "automatic-plot-ia",
        "automaticplotia",
        "autoplotia",
        "auto-plot-ia",
    ]:
        args.action = "plot-automatic-ia"

    if args.action in [
        "description",
        "plot-automatic",
        "plot-automatic-ia",
    ]:
        req_files = get_file_paths(
            path_to_lobster_calc=Path(os.getcwd()), requested_files=["structure", "charge", "icohplist", "cohpcar"]
        )

        if args.coops:
            req_files_coops = get_file_paths(
                path_to_lobster_calc=Path(os.getcwd()), requested_files=["icooplist", "coopcar"]
            )

            req_files["icohplist"] = req_files_coops["icooplist"]
            req_files["cohpcar"] = req_files_coops["coopcar"]

        if args.cobis:
            req_files_cobis = get_file_paths(
                path_to_lobster_calc=Path(os.getcwd()), requested_files=["icobilist", "cobicar"]
            )

            req_files["icohplist"] = req_files_cobis["icobilist"]
            req_files["cohpcar"] = req_files_cobis["cobicar"]

        for arg_name in req_files:
            setattr(args, arg_name, req_files[arg_name])

    if args.action in ["description", "plot-automatic", "plot-automatic-ia"]:
        which_bonds = "all" if args.allbonds else "cation-anion"

        analyse = Analysis(
            path_to_poscar=args.structure,
            path_to_charge=args.charge,
            path_to_cohpcar=args.cohpcar,
            path_to_icohplist=args.icohplist,
            which_bonds=which_bonds,
            are_coops=args.coops,
            are_cobis=args.cobis,
            noise_cutoff=args.noisecutoff,
            cutoff_icohp=args.cutofficohp,
            orbital_cutoff=args.orbitalcutoff,
            orbital_resolved=args.orbitalresolved,
            summed_spins=args.summedspins,
        )

        describe = Description(analysis_object=analyse)
        describe.write_description()

        if args.file_json is not None:
            analysedict = analyse.condensed_bonding_analysis
            with open(args.file_json, "w") as fd:
                json.dump(analysedict, fd)

    if args.action in [
        "plot",
        "plot-automatic",
        "plot-automatic-ia",
        "plot-dos",
        "plotdos",
        "plot-icohp-distance",
        "ploticohpdistance",
    ]:
        style_kwargs = {}
        style_kwargs.update(_user_figsize(args.width, args.height))
        if args.fontsize:
            style_kwargs.update({"font.size": args.fontsize})

        style_list = get_style_list(no_base_style=args.no_base_style, styles=args.style, **style_kwargs)
        matplotlib.style.use(style_list)

        if args.sigma:
            sigma = args.sigma
        elif args.fwhm:
            sigma = args.fwhm / (2 * sqrt(2 * log(2)))
        else:
            sigma = None

    if args.action in ["plot-automatic"]:
        describe.plot_cohps(
            ylim=args.ylim,
            xlim=args.xlim,
            integrated=args.integrated,
            save=args.save_plot is not None,
            filename=args.save_plot,
            title=args.title,
            sigma=sigma,
            hide=args.hideplot,
        )

    if args.action in ["plot-automatic-ia"]:
        describe.plot_interactive_cohps(
            ylim=args.ylim,
            xlim=args.xlim,
            integrated=args.integrated,
            save_as_html=args.save_plot,
            filename=args.save_plot,
            title=args.title,
            sigma=sigma,
            hide=args.hideplot,
            label_resolved=args.labelresolved,
            orbital_resolved=args.orbitalresolved,
        )

    if args.action == "plot":
        if args.cobis:
            filename = get_file_paths(path_to_lobster_calc=Path(os.getcwd()), requested_files=["cobicar"]).get(
                "cobicar"
            )
            options = {"are_cobis": True, "are_coops": False}
        elif args.multi_cobis:
            filename = get_file_paths(path_to_lobster_calc=Path(os.getcwd()), requested_files=["cobicar"]).get(
                "cobicar"
            )
            options = {"are_cobis": False, "are_coops": False, "are_multi_center_cobis": True}
        elif args.coops:
            filename = get_file_paths(path_to_lobster_calc=Path(os.getcwd()), requested_files=["coopcar"]).get(
                "coopcar"
            )
            options = {"are_cobis": False, "are_coops": True}
        else:
            filename = get_file_paths(path_to_lobster_calc=Path(os.getcwd()), requested_files=["cohpcar"]).get(
                "cohpcar"
            )
            options = {"are_cobis": False, "are_coops": False}

        struture_filename = get_file_paths(path_to_lobster_calc=Path(os.getcwd()), requested_files=["structure"]).get(
            "structure"
        )

        completecohp = CompleteCohp.from_file(
            fmt="LOBSTER",
            filename=filename,
            structure_file=struture_filename,
            **options,
        )
        if args.multi_cobis:
            options = {"are_cobis": True, "are_coops": False}
        cp = PlainCohpPlotter(**options)

        if not args.summed:
            # TODO: add checks for label in allowed labels -> print all labels
            # TODO: add check if args.oribtalwise is exactly as long as labels
            # TODO: add check if orbital is in args.orbitalwise

            for label in args.bond_numbers:
                if str(label) not in completecohp.bonds:
                    raise IndexError(
                        "The provided bond label " + str(label) + " is not available in ICO**LIST.lobster.\n "
                        "Allowed options are in this list: \n"
                        + str([int(listi) for listi in list(completecohp.bonds.keys())])
                    )

            if not args.orbitalwise:
                for label in args.bond_numbers:
                    cp.add_cohp(label, completecohp.get_cohp_by_label(label=str(label)))
            else:
                if len(args.bond_numbers) != len(args.orbitalwise):
                    raise IndexError(
                        "Please provide as mainy orbitals as bond labels,"
                        " e.g., lobsterpy plot 1 1 --orbitalwise '2s-2s' '2s-2px'"
                    )

                for ilabel, label in enumerate(args.bond_numbers):
                    orbitals = args.orbitalwise[ilabel]

                    availableorbitals = list(completecohp.orb_res_cohp[str(label)].keys())
                    orbitaloptions = [*availableorbitals, "all"]

                    if orbitals not in orbitaloptions:
                        raise IndexError(
                            "Orbital in not available for current bond. \n"
                            "For bond "
                            + str(label)
                            + " only the following orbital options are available: \n"
                            + str(orbitaloptions)
                        )

                    if orbitals != "all":
                        cp.add_cohp(
                            str(label) + ": " + orbitals,
                            completecohp.get_orbital_resolved_cohp(label=str(label), orbitals=orbitals),
                        )
                    else:
                        for orbitals in availableorbitals:
                            cp.add_cohp(
                                str(label) + ": " + orbitals,
                                completecohp.get_orbital_resolved_cohp(label=str(label), orbitals=orbitals),
                            )
        else:
            cp.add_cohp(
                str(args.bond_numbers),
                completecohp.get_summed_cohp_by_label_list(label_list=[str(label) for label in args.bond_numbers]),
            )

        plt = cp.get_plot(integrated=args.integrated, xlim=args.xlim, ylim=args.ylim, sigma=sigma)

        ax = plt.gca()
        ax.set_title(args.title)

        if not args.hideplot and not args.save_plot:
            plt.show()
        elif args.save_plot and not args.hideplot:
            fig = plt.gcf()
            fig.savefig(args.save_plot)
            plt.show()
        if args.save_plot and args.hideplot:
            plt.savefig(args.save_plot)

    if args.action in ["create-inputs", "createinputs"]:
        from pymatgen.core.structure import Structure
        from pymatgen.io.lobster import Lobsterin

        # Check for .gz files exist for default values and update accordingly
        default_files = {
            "structure": "POSCAR",
            "potcar": "POTCAR",
            "incar": "INCAR",
        }

        for arg_name in default_files:
            file_path = getattr(args, arg_name)
            if not file_path.exists():
                gz_file_path = file_path.with_name(zpath(file_path.name))
                if gz_file_path.exists():
                    setattr(args, arg_name, gz_file_path)
                else:
                    raise ValueError(
                        "Files necessary for creating inputs for LOBSTER calcs not found in the current directory."
                    )

        if args.userbasis is None:
            # This will rely on standard basis files as stored in pymatgen

            potcar_names = Lobsterin._get_potcar_symbols(POTCAR_input=args.potcar)

            list_basis_dict = Lobsterin.get_all_possible_basis_functions(
                structure=Structure.from_file(args.structure),
                potcar_symbols=potcar_names,
            )

            for ibasis, basis_dict in enumerate(list_basis_dict):
                lobsterinput = Lobsterin.standard_calculations_from_vasp_files(
                    args.structure,
                    args.incar,
                    None,
                    option="standard",
                    dict_for_basis=basis_dict,
                )

                lobsterin_path = Path(str(args.lobsterinout) + "-" + str(ibasis))
                incar_path = Path(str(args.incarout) + "-" + str(ibasis))

                if (not lobsterin_path.is_file() and not incar_path.is_file()) or (args.overwrite):
                    lobsterinput.write_lobsterin(lobsterin_path)
                    lobsterinput.write_INCAR(
                        incar_input=args.incar,
                        incar_output=incar_path,
                        poscar_input=args.structure,
                        isym=0,
                    )
                else:
                    raise ValueError('please use "--overwrite" if you would like to overwrite existing lobster inputs')
        else:
            # convert list userbasis to dict
            userbasis = {}
            for userbasis_single in args.userbasis:
                userbasis[userbasis_single[0]] = userbasis_single[1]

            lobsterinput = Lobsterin.standard_calculations_from_vasp_files(
                args.structure,
                args.incar,
                None,
                option="standard",
                dict_for_basis=userbasis,
            )

            lobsterin_path = Path(str(args.lobsterinout) + "-" + str(0))
            incar_path = Path(str(args.incarout) + "-" + str(0))

            if (not lobsterin_path.is_file() and not incar_path.is_file()) or (args.overwrite):
                lobsterinput.write_lobsterin(lobsterin_path)
                lobsterinput.write_INCAR(
                    incar_input=args.incar,
                    incar_output=incar_path,
                    poscar_input=args.structure,
                    isym=0,
                )
            else:
                raise ValueError('please use "--overwrite" if you would like to overwrite existing lobster inputs')

    if args.action in ["description-quality"]:
        # Check for .gz files exist for default values and update accordingly
        req_files = get_file_paths(
            path_to_lobster_calc=Path(os.getcwd()), requested_files=["structure", "lobsterin", "lobsterout"]
        )
        for arg_name in req_files:
            setattr(args, arg_name, req_files[arg_name])

        optional_files = {
            "bandoverlaps": "bandOverlaps.lobster",
            "potcar": "POTCAR",
        }

        for arg_name in optional_files:
            file_path = getattr(args, arg_name)
            if not file_path.exists():
                gz_file_path = file_path.with_name(zpath(file_path.name))
                if gz_file_path.exists():
                    setattr(args, arg_name, gz_file_path)

        bva_comp = args.bvacomp

        if bva_comp:
            bva_files = get_file_paths(path_to_lobster_calc=Path(os.getcwd()), requested_files=["charge"])
            for arg_name in bva_files:
                setattr(args, arg_name, bva_files[arg_name])

        dos_comparison = args.doscomp

        if dos_comparison:
            if "DOSCAR.LSO.lobster" in args.doscar.name:
                dos_files = get_file_paths(
                    path_to_lobster_calc=Path(os.getcwd()), requested_files=["vasprun", "doscar"], use_lso_dos=True
                )
            else:
                dos_files = get_file_paths(
                    path_to_lobster_calc=Path(os.getcwd()), requested_files=["vasprun", "doscar"]
                )
            for arg_name in dos_files:
                setattr(args, arg_name, dos_files[arg_name])

        potcar_file_path = args.potcar

        quality_dict = Analysis.get_lobster_calc_quality_summary(
            path_to_poscar=args.structure,
            path_to_charge=args.charge,
            path_to_lobsterout=args.lobsterout,
            path_to_lobsterin=args.lobsterin,
            path_to_potcar=None if not potcar_file_path.exists() else potcar_file_path,
            potcar_symbols=args.potcarsymbols,
            path_to_bandoverlaps=args.bandoverlaps,
            dos_comparison=dos_comparison,
            bva_comp=bva_comp,
            path_to_doscar=args.doscar,
            e_range=args.erange,
            n_bins=args.nbins,
            path_to_vasprun=args.vasprun,
        )

        quality_text = Description.get_calc_quality_description(quality_dict)
        Description.write_calc_quality_description(quality_text)

        if args.file_calc_quality_json is not None:
            with open(args.file_calc_quality_json, "w") as fd:
                json.dump(quality_dict, fd)

    if args.action in ["plot-dos", "plotdos"]:
        if "DOSCAR.LSO.lobster" in args.doscar.name:
            req_files = get_file_paths(
                path_to_lobster_calc=Path(os.getcwd()), requested_files=["structure", "doscar"], use_lso_dos=True
            )
        else:
            req_files = get_file_paths(path_to_lobster_calc=Path(os.getcwd()), requested_files=["structure", "doscar"])

        for arg_name in req_files:
            setattr(args, arg_name, req_files[arg_name])

        from pymatgen.io.lobster import Doscar

        lobs_dos = Doscar(doscar=args.doscar, structure_file=args.structure).completedos

        dos_plotter = PlainDosPlotter(summed=args.summedspins, sigma=args.sigma)
        if args.addtotal:
            dos_plotter.add_dos(dos=lobs_dos, label="Total DOS")
        if args.spddos:
            dos_plotter.add_dos_dict(dos_dict=lobs_dos.get_spd_dos())

        if args.elementdos:
            dos_plotter.add_dos_dict(dos_dict=lobs_dos.get_element_dos())

        if args.element:
            for element in args.element:
                element_spddos = lobs_dos.get_element_spd_dos(el=element)
                for orbital, dos in element_spddos.items():
                    label = f"{element}: {orbital.name}"
                    dos_plotter.add_dos_dict(dos_dict={label: dos})

        if args.site is not None and args.orbital:
            if len(args.site) > len(args.orbital):
                for site in args.site:
                    for orbital in args.orbital:
                        dos_plotter.add_site_orbital_dos(site_index=site, orbital=orbital, dos=lobs_dos)
            elif len(args.orbital) > len(args.site):
                for orbital in args.orbital:
                    for site in args.site:
                        dos_plotter.add_site_orbital_dos(site_index=site, orbital=orbital, dos=lobs_dos)
            else:
                for site, orbital in zip(args.site, args.orbital):
                    dos_plotter.add_site_orbital_dos(site_index=site, orbital=orbital, dos=lobs_dos)
        elif (
            args.site is None
            and not args.orbital
            and not args.element
            and not args.spddos
            and not args.elementdos
            and not args.addtotal
        ):
            dos_plotter.add_dos(dos=lobs_dos, label="Total DOS")
            dos_plotter.add_dos_dict(dos_dict=lobs_dos.get_element_dos())

        elif (args.site is None or not args.orbital) and (not args.element and not args.spddos and not args.elementdos):
            raise ValueError("Please set both args i.e site and orbital to generate the plot")

        plt = dos_plotter.get_plot(
            xlim=args.xlim,
            ylim=args.ylim,
            beta_dashed=True,
            invert_axes=args.invertaxis,
        )

        ax = plt.gca()
        ax.set_title(args.title)

        if not args.hideplot and not args.save_plot:
            plt.show()
        elif args.save_plot and not args.hideplot:
            fig = plt.gcf()
            fig.savefig(args.save_plot)
            plt.show()
        if args.save_plot and args.hideplot:
            plt.savefig(args.save_plot)

    if args.action in ["plot-icohp-distance", "ploticohpdistance"]:
        if args.cobis:
            filename = get_file_paths(path_to_lobster_calc=Path(os.getcwd()), requested_files=["icobilist"]).get(
                "icobilist"
            )
            args.icohplist = filename
            # filename = args.icohplist.parent / "ICOBILIST.lobster"
            # if not filename.exists():
            #     filename = filename.with_name(zpath(filename.name))
            options = {"are_cobis": True, "are_coops": False}
        elif args.coops:
            filename = get_file_paths(path_to_lobster_calc=Path(os.getcwd()), requested_files=["icooplist"]).get(
                "icooplist"
            )
            args.icohplist = filename
            options = {"are_cobis": False, "are_coops": True}
        else:
            filename = get_file_paths(path_to_lobster_calc=Path(os.getcwd()), requested_files=["icohplist"]).get(
                "icohplist"
            )
            args.icohplist = filename
            options = {"are_cobis": False, "are_coops": False}

        icohpcollection = Icohplist(filename=args.icohplist, **options).icohpcollection
        icohp_plotter = IcohpDistancePlotter(**options)

        icohp_plotter.add_icohps(icohpcollection=icohpcollection, label="")

        plt = icohp_plotter.get_plot(xlim=args.xlim, ylim=args.ylim)

        ax = plt.gca()
        ax.set_title(args.title)

        if not args.hideplot and not args.save_plot:
            plt.show()
        elif args.save_plot and not args.hideplot:
            fig = plt.gcf()
            fig.savefig(args.save_plot)
            plt.show()
        if args.save_plot and args.hideplot:
            plt.savefig(args.save_plot)


if __name__ == "__main__":
    main()



# Copyright (c) lobsterpy development team
# Distributed under the terms of a BSD 3-Clause "New" or "Revised" License

"""This module provides a class for generating graph objects using lobsterpy data."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

from pymatgen.core.structure import Structure
from pymatgen.io.lobster.lobsterenv import LobsterNeighbors
from pymatgen.io.lobster.outputs import Charge

from lobsterpy.cohp.analyze import Analysis


class LobsterGraph:
    """
    Class to generate structure graph objects with bonding data from Lobster.

    Mode of automatic bonding analysis can be `cation-anion` or `all` bonds. The strongest bond is
    determined based on the ICOHPs. The coordination environments are determined based on
    `cutoff_icohp * ICOHPs` values. If the path of ICOBILIST (ICOOPLIST) is provided, the ICOBI (ICOOP)
    values corresponding to relevant bond labels obtained from the ICOHPLIST are also added as edge properties
    to the structure graph objects. The Mulliken and Loewdin charges are added as node properties to
    the structure graph objects.

    Attributes:
        sg: return structure_graph object

    :param path_to_poscar: path to POSCAR (e.g., "POSCAR").
    :param path_to_charge: path to CHARGE.lobster (e.g., "CHARGE.lobster").
    :param path_to_cohpcar: path to COHPCAR.lobster (e.g., "COHPCAR.lobster").
    :param path_to_icohplist: path to ICOHPLIST.lobster (e.g., "ICOHPLIST.lobster").
    :param path_to_icooplist: path to ICOOPLIST.lobster (e.g., "ICOOPLIST.lobster").
    :param path_to_icobilist: path to ICOBILIST.lobster (e.g., "ICOBILIST.lobster").
    :param path_to_madelung: path to MadelungEnergies.lobster (e.g., "MadelungEnergies.lobster")
    :param cutoff_icohp: only bonds that are stronger than cutoff_icohp * strongest ICOHP will be considered.
    :param noise_cutoff: if provided hardcodes the lower limit of icohps considered.
    :param add_additional_data_sg: if True will add the information from ICOOPLIST.lobster
        and ICOBILIST.lobster based on ICOHPLIST.lobster relevant bond.
    :param which_bonds: selects which kind of bonds are analyzed. "all" is the default.
    :param start: start energy for bonding antibonding percent integration.
    """

    def __init__(
        self,
        path_to_poscar: str | Path,
        path_to_charge: str | Path,
        path_to_cohpcar: str | Path,
        path_to_icohplist: str | Path,
        path_to_madelung: str | Path,
        add_additional_data_sg: bool = True,
        path_to_icooplist: str | Path | None = None,
        path_to_icobilist: str | Path | None = None,
        which_bonds: str = "all",
        cutoff_icohp: float = 0.10,
        noise_cutoff: float = 0.1,
        start: float | None = None,
    ):
        """Initialize and return a structure graph object."""
        if add_additional_data_sg:
            self.add_additional_data_sg = add_additional_data_sg
            if path_to_icooplist is not None and path_to_icobilist is not None:
                self.path_to_icooplist = path_to_icooplist
                self.path_to_icobilist = path_to_icobilist
            else:
                raise ValueError(
                    "add_additional_data_sg is set to True.Please provide path_to_icooplist and path_to_icobilist"
                )
        else:
            self.add_additional_data_sg = add_additional_data_sg  # type: ignore

        self.path_to_poscar = path_to_poscar
        self.path_to_charge = path_to_charge
        self.path_to_cohpcar = path_to_cohpcar
        self.path_to_icohplist = path_to_icohplist
        self.path_to_madelung = path_to_madelung
        self.which_bonds = which_bonds
        self.cutoff_icohp = cutoff_icohp
        self.noise_cutoff = noise_cutoff

        if self.which_bonds == "all":
            self.additional_condition = 0
        elif self.which_bonds == "cation-anion":
            self.additional_condition = 1
        else:
            raise ValueError(
                "Only accepted values are 'all' and 'cation-anion'."
                "Please check the input parameters of which_bonds arg"
            )
        self.start = start

        self.sg = self.get_decorated_sg()

    def get_decorated_sg(self):
        """
        Create a graph object decorated with bonding data from LobsterPy.

        Returns:
            structure graph object
        """
        if self.add_additional_data_sg:
            chemenvlobster = LobsterNeighbors(
                are_coops=False,
                filename_icohp=self.path_to_icohplist,
                perc_strength_icohp=self.cutoff_icohp,
                structure=Structure.from_file(self.path_to_poscar),
                additional_condition=self.additional_condition,
                filename_charge=self.path_to_charge,
                add_additional_data_sg=self.add_additional_data_sg,
                filename_blist_sg1=self.path_to_icobilist,
                id_blist_sg1="ICOBI",
                filename_blist_sg2=self.path_to_icooplist,
                id_blist_sg2="ICOOP",
                valences_from_charges=True,
                adapt_extremum_to_add_cond=True,
                noise_cutoff=self.noise_cutoff,
            )

        else:
            chemenvlobster = LobsterNeighbors(
                are_coops=False,
                filename_icohp=self.path_to_icohplist,
                perc_strength_icohp=self.cutoff_icohp,
                structure=Structure.from_file(self.path_to_poscar),
                additional_condition=self.additional_condition,
                filename_charge=self.path_to_charge,
                add_additional_data_sg=self.add_additional_data_sg,
                valences_from_charges=True,
                adapt_extremum_to_add_cond=True,
                noise_cutoff=self.noise_cutoff,
            )

        # Adds Mulliken and Löwdin charges as site properties to structure object (node properties)
        decorated_structure = Charge(self.path_to_charge).get_structure_with_charges(self.path_to_poscar)

        # Create the structure graph object decorated with site and edge properties based on ICOHP/ICOBI/ICOOP data
        lobster_env = chemenvlobster.get_bonded_structure(
            structure=decorated_structure, decorate=True, edge_properties=True
        )

        # Initialize automating bonding analysis from Lobsterpy based on ICOHP
        analyze = Analysis(
            path_to_charge=self.path_to_charge,
            path_to_cohpcar=self.path_to_cohpcar,
            path_to_poscar=self.path_to_poscar,
            path_to_icohplist=self.path_to_icohplist,
            path_to_madelung=self.path_to_madelung,
            which_bonds=self.which_bonds,
            cutoff_icohp=self.cutoff_icohp,
        )

        # Store the summarized dictionary object containing bonding information
        cba = analyze.condensed_bonding_analysis

        # Iterate over sites in the dictionary
        for cba_data in cba["sites"].values():
            for _, node_data in lobster_env.graph.nodes.data():
                # Check if ions are same and add its corresponding environment in node properties
                if cba_data["ion"] == node_data["specie"]:
                    node_data["properties"].update({"env": cba_data["env"]})

        for edge_prop in lobster_env.graph.edges.data():  # Iterate over structure graph edges
            _ab, ab_p, _b, b_p = analyze._integrate_antbdstates_below_efermi(
                cohp=analyze.chemenv.completecohp.get_cohp_by_label(edge_prop[2]["bond_label"]),
                start=self.start,
            )  # Compute bonding- antibonding percentages for each bond in structure graph object
            edge_prop[2]["ICOHP_bonding_perc"] = b_p  # Store bonding percentage in edge of graph object
            edge_prop[2]["ICOHP_antibonding_perc"] = ab_p  # Store anti-bonding percentage in edge graph object

        return lobster_env
